"""Schema for the error objects returned by the validation command"""
from pydantic import BaseModel, Field, constr
from typing import List, Union
from enum import Enum

from .geometry import Point3D, LineSegment3D


class ExtensionTypes(str, Enum):
    """Types of Honeybee/Dragonfly extensions."""
    core = 'Core'
    radiance = 'Radiance'
    energy = 'Energy'


class ObjectTypes(str, Enum):
    """Types of Honeybee objects."""
    shade = 'Shade'
    aperture = 'Aperture'
    door = 'Door'
    sub_face = 'SubFace'
    face = 'Face'
    room = 'Room'
    sensor_grid = 'SensorGrid'
    view = 'View'
    modifier = 'Modifier'
    modifier_set = 'ModifierSet'
    material = 'Material'
    construction = 'Construction'
    construction_set = 'ConstructionSet'
    schedule_type_limit = 'ScheduleTypeLimit'
    schedule = 'Schedule'
    program_type = 'ProgramType'
    hvac = 'HVAC'
    shw = 'SHW'
    roof_spec = 'RoofSpecification'
    room_2d = 'Room2D'
    story = 'Story'
    building = 'Building'


class ParentTypes(str, Enum):
    """Types of Honeybee objects that can be parents."""
    aperture = 'Aperture'
    door = 'Door'
    face = 'Face'
    room = 'Room'
    story = 'Story'
    building = 'Building'


class ValidationParent(BaseModel):

    type: constr(regex='^ValidationParent$') = 'ValidationParent'

    parent_type: ParentTypes = Field(
        ...,
        description='Text for the type of object that the parent is.'
    )

    id: str = Field(
        ...,
        regex=r'^[.A-Za-z0-9_-]+$',
        min_length=1,
        max_length=100,
        description='Text string for the unique ID of the parent object.'
    )

    name: str = Field(
        default=None,
        description='Display name of the parent object.'
    )


class ValidationError(BaseModel):

    type: constr(regex='^ValidationError$') = 'ValidationError'

    code: str = Field(
        ...,
        min_length=6,
        max_length=6,
        regex=r'([0-9]+)',
        description='Text with 6 digits for the error code. The first two digits '
        'indicate whether the error is a core honeybee error (00) vs. an extension '
        'error (any non-zero number). The second two digits indicate the nature '
        'of the error (00 is an identifier error, 01 is a geometry error, 02 is an '
        'adjacency error). The third two digits are used to give a unique ID to '
        'each condition moving upwards from more specific/detailed objects/errors '
        'to coarser/more abstract objects/errors. A full list of error codes can '
        'be found here: https://docs.pollination.solutions/user-manual/rhino-plugin/'
        'troubleshooting/help-with-modeling-error-codes'
    )

    error_type: str = Field(
        ...,
        description='A human-readable version of the error code, typically not more '
        'than five words long.'
    )

    extension_type: ExtensionTypes = Field(
        ...,
        description='Text for the Honeybee extension from which the error originated '
        '(from the ExtensionTypes enumeration).'
    )

    element_type: ObjectTypes = Field(
        ...,
        description='Text for the type of object that caused the error.'
    )

    element_id: List[
        constr(min_length=1, max_length=100, regex=r'^[^,;!\n\t]+$')
    ] = Field(
        ...,
        description='A list of text strings for the unique object IDs that caused '
        'the error. The list typically contains a single item but there are some types '
        'errors that stem from multiple objects like mis-matched area adjacencies '
        'or overlapping Room geometries. Note that the IDs in this list can be the '
        'identifier of a core object like a Room or a Face or it can be for an '
        'extension object like a SensorGrid or a Construction.'
    )

    element_name: List[str] = Field(
        default=None,
        description='A list of text strings for the display names of the objects '
        'that caused the error.'
    )

    message: str = Field(
        ...,
        description='Text for the error message with a detailed description of '
        'what exactly is invalid about the element.'
    )

    parents: List[List[ValidationParent]] = Field(
        default=None,
        description='A list lists where each sub-list corresponds to one of the '
        'objects in the element_id property. Each sub-list contains information for '
        'the parent objects of the object that caused the error. '
        'This can be useful for locating the problematic object in the model. '
        'This will contain 1 item for a Face with a parent Room. It will contain 2 '
        'for an Aperture that has a parent Face with a parent Room.'
    )

    top_parents: List[ValidationParent] = Field(
        default=None,
        description='A list of top-level parent objects for the specific case of '
        'duplicate child-object identifiers, where several top-level parents '
        'are involved.'
    )

    helper_geometry: List[Union[Point3D, LineSegment3D]] = Field(
        default=None,
        description='An optional list of geometry objects that helps illustrate '
        'where exactly issues with invalid geometry exist within the Honeybee object. '
        'Examples include the naked and non-manifold line segments for non-solid Room '
        'geometries, the points of self-intersection for cases of self-intersecting '
        'geometry and out-of-plane vertices for non-planar objects. Oftentimes, '
        'zooming directly to these helper geometries will help end users understand '
        'invalid situations in their model faster than simple zooming to the invalid '
        'Honeybee object in its totality.'
    )


class ValidationReport(BaseModel):

    type: constr(regex='^ValidationReport$') = 'ValidationReport'

    app_name: str = Field(
        'Honeybee',
        description='Text string for the name of the application that performed '
        'the validation. This is typically either Honeybee or Dragonfly.'
    )

    app_version: str = Field(
        ...,
        regex=r'([0-9]+)\.([0-9]+)\.([0-9]+)',
        description='Text string for the version of honeybee-core or dragonfly-core '
        'that performed the validation.'
    )

    schema_version: str = Field(
        ...,
        regex=r'([0-9]+)\.([0-9]+)\.([0-9]+)',
        description='Text string for the version of honeybee-schema or dragonfly-schema '
        'that performed the validation.'
    )

    valid: bool = Field(
        ...,
        description='Boolean to note whether the Model is valid or not.'
    )

    fatal_error: str = Field(
        '',
        description='A text string containing an exception if the Model failed to '
        'serialize. It will be an empty string if serialization was successful.'
    )

    errors: List[ValidationError] = Field(
        default=None,
        description='A list of objects for each error that was discovered in the model. '
        'This will be an empty list or None if no errors were found.'
    )



"""generate openapi docs."""
from pkg_resources import get_distribution
from pydantic_openapi_helper.core import get_openapi
from pydantic_openapi_helper.inheritance import class_mapper
from honeybee_schema.model import Model
from honeybee_schema.energy.simulation import SimulationParameter
from honeybee_schema.validation import ValidationReport
from honeybee_schema.comparison import ComparisonReport, SyncInstructions
from honeybee_schema.projectinfo import ProjectInfo

import json
import argparse

parser = argparse.ArgumentParser(description='Generate OpenAPI JSON schemas')

parser.add_argument('--version', help='Set the version of the new OpenAPI Schema')

args = parser.parse_args()

if args.version:
    VERSION = args.version.replace('v', '')
else:
    VERSION = '.'.join(get_distribution('honeybee_schema').version.split('.')[:3])

info = {
    "description": "",
    "version": VERSION,
    "title": "",
    "contact": {
        "name": "Ladybug Tools",
        "email": "info@ladybug.tools",
        "url": "https://github.com/ladybug-tools/honeybee-schema"
    },
    "x-logo": {
        "url": "https://www.ladybug.tools/assets/img/honeybee-large.png",
        "altText": "Honeybee logo"
    },
    "license": {
        "name": "BSD",
        "url": "https://github.com/ladybug-tools-in2/honeybee-schema/blob/master/LICENSE"
    }
}

modules = [
    {'module': [Model], 'name': 'Model'},
    {'module': [SimulationParameter], 'name': 'Simulation Parameter'},
    {'module': [ValidationReport], 'name': 'Validation Report'},
    {'module': [ComparisonReport], 'name': 'Comparison Report'},
    {'module': [SyncInstructions], 'name': 'Sync Instructions'},
    {'module': [ProjectInfo], 'name': 'Project Information'}
]


def _process_name(name):
    """Process module name."""
    new_name = '-'.join(n.lower() for n in name.split())
    return new_name


for module in modules:
    # generate Recipe open api schema
    print(f'Generating {module["name"]} documentation...')

    external_docs = {
        "description": "OpenAPI Specification with Inheritance",
        "url": f"./{_process_name(module['name'])}_inheritance.json"
    }

    openapi = get_openapi(
        module['module'],
        title=f'Honeybee {module["name"]} Schema',
        description=f'Honeybee {_process_name(module["name"])} schema.',
        version=VERSION, info=info,
        external_docs=external_docs
    )

    # Make Model version read-only
    if module['module'][0] is Model:
        model_object = openapi['components']['schemas']['Model']
        model_object['properties']['version']['readOnly'] = True
        model_object['properties']['version']['default'] = VERSION

    with open(f'./docs/{_process_name(module["name"])}.json', 'w') as out_file:
        json.dump(openapi, out_file, indent=2)

    # with inheritance
    openapi = get_openapi(
        module['module'],
        title=f'Honeybee {module["name"]} Schema',
        description=f'Documentation for Honeybee {_process_name(module["name"])} schema',
        version=VERSION, info=info,
        inheritance=True,
        external_docs=external_docs
    )

    # set the version default key in the Recipe schema
    if module['module'][0] is Model:
        model_object = openapi['components']['schemas']['Model']
        model_object['allOf'][1]['properties']['version']['readOnly'] = True
        model_object['allOf'][1]['properties']['version']['default'] = VERSION

    with open(f'./docs/{_process_name(module["name"])}_inheritance.json', 'w') \
            as out_file:
        json.dump(openapi, out_file, indent=2)

    # add the mapper file
    with open(f'./docs/{_process_name(module["name"])}_mapper.json', 'w') as out_file:
        json.dump(class_mapper(module['module']), out_file, indent=2)

# generate JSONSchema for Honeybee model
with open('./docs/model_json_schema.json', 'w') as out_file:
    out_file.write(Model.schema_json(indent=2))

# generate schema for mode with inheritance but without discriminator
# we will use this file for generating redocly - the full model is too big, and the
# model with inheritance and discriminators is renders incorrectly
external_docs = {
    "description": "OpenAPI Specification with Inheritance",
    "url": "./model_inheritance.json"
}

openapi = get_openapi(
    [Model],
    title='Honeybee Model Schema',
    description='Documentation for Honeybee model schema',
    version=VERSION, info=info,
    inheritance=True,
    external_docs=external_docs,
    add_discriminator=False
)

with open('./docs/model_redoc.json', 'w') as out_file:
    json.dump(openapi, out_file, indent=2)



"""All-air HVAC systems, providing ventilation and meeting thermal demand with air."""
from pydantic import Field, constr
from enum import Enum

from ._template import _TemplateSystem


class AllAirEconomizerType(str, Enum):
    no_economizer = 'NoEconomizer'
    differential_dry_bulb = 'DifferentialDryBulb'
    differential_enthalpy = 'DifferentialEnthalpy'
    differential_dry_bulb_and_enthalpy = 'DifferentialDryBulbAndEnthalpy'
    fixed_dry_bulb = 'FixedDryBulb'
    fixed_enthalpy = 'FixedEnthalpy'
    electronic_enthalpy = 'ElectronicEnthalpy'


class _AllAirBase(_TemplateSystem):
    """Base class for all-air systems.

    All-air systems provide both ventilation and heating + cooling demand with
    the same stream of warm/cool air. As such, they often grant tight control
    over zone humidity. However, because such systems often involve the
    cooling of air only to reheat it again, they are often more energy intensive
    than systems that separate ventilation from the meeting of thermal loads.
    """

    economizer_type: AllAirEconomizerType = Field(
        AllAirEconomizerType.no_economizer,
        description='Text to indicate the type of air-side economizer used on '
        'the system (from the AllAirEconomizerType enumeration).'
    )

    sensible_heat_recovery: float = Field(
        0,
        ge=0,
        le=1,
        description='A number between 0 and 1 for the effectiveness of sensible '
        'heat recovery within the system.'
    )

    latent_heat_recovery: float = Field(
        0,
        ge=0,
        le=1,
        description='A number between 0 and 1 for the effectiveness of latent '
        'heat recovery within the system.'
    )

    demand_controlled_ventilation: bool = Field(
        False,
        description='Boolean to note whether demand controlled ventilation should be '
        'used on the system, which will vary the amount of ventilation air according '
        'to the occupancy schedule of the Rooms.'
    )


class VAVEquipmentType(str, Enum):
    vav_chill_gbr = 'VAV_Chiller_Boiler'
    vav_chill_ashp = 'VAV_Chiller_ASHP'
    vav_chill_dhw = 'VAV_Chiller_DHW'
    vav_chill_pfp = 'VAV_Chiller_PFP'
    vav_chill_gcr = 'VAV_Chiller_GasCoil'
    vav_ac_chill_gbr = 'VAV_ACChiller_Boiler'
    vav_ac_chill_ashp = 'VAV_ACChiller_ASHP'
    vav_ac_chill_dhw = 'VAV_ACChiller_DHW'
    vav_ac_chill_pfp = 'VAV_ACChiller_PFP'
    vav_ac_chill_gcr = 'VAV_ACChiller_GasCoil'
    vav_dcw_gbr = 'VAV_DCW_Boiler'
    vav_dcw_ashp = 'VAV_DCW_ASHP'
    vav_dcw_dhw = 'VAV_DCW_DHW'
    vav_dcw_pfp = 'VAV_DCW_PFP'
    vav_dcw_gcr = 'VAV_DCW_GasCoil'


class PVAVEquipmentType(str, Enum):
    pvav_gbr = 'PVAV_Boiler'
    pvav_ashp = 'PVAV_ASHP'
    pvav_dhw = 'PVAV_DHW'
    pvav_pfp = 'PVAV_PFP'
    pvav_ger = 'PVAV_BoilerElectricReheat'


class PSZEquipmentType(str, Enum):
    psz_e_base = 'PSZAC_ElectricBaseboard'
    psz_gb_base = 'PSZAC_BoilerBaseboard'
    psz_dhw_base = 'PSZAC_DHWBaseboard'
    psz_guh = 'PSZAC_GasHeaters'
    psz_ec = 'PSZAC_ElectricCoil'
    psz_gc = 'PSZAC_GasCoil'
    psz_gb = 'PSZAC_Boiler'
    psz_ashp = 'PSZAC_ASHP'
    psz_dhw = 'PSZAC_DHW'
    psz_ac = 'PSZAC'
    psz_dcw_e_base = 'PSZAC_DCW_ElectricBaseboard'
    psz_dcw_gb_base = 'PSZAC_DCW_BoilerBaseboard'
    psz_dcw_guh = 'PSZAC_DCW_GasHeaters'
    psz_dcw_ec = 'PSZAC_DCW_ElectricCoil'
    psz_dcw_gc = 'PSZAC_DCW_GasCoil'
    psz_dcw_gb = 'PSZAC_DCW_Boiler'
    psz_dcw_ashp = 'PSZAC_DCW_ASHP'
    psz_dcw_dhw = 'PSZAC_DCW_DHW'
    psz_dcw_ac = 'PSZAC_DCW'
    psz_hp = 'PSZHP'


class PTACEquipmentType(str, Enum):
    ptac_e_base = 'PTAC_ElectricBaseboard'
    ptac_gb_base = 'PTAC_BoilerBaseboard'
    ptac_dhw_base = 'PTAC_DHWBaseboard'
    ptac_guh = 'PTAC_GasHeaters'
    ptac_ec = 'PTAC_ElectricCoil'
    ptac_gc = 'PTAC_GasCoil'
    ptac_gb = 'PTAC_Boiler'
    ptac_ashp = 'PTAC_ASHP'
    ptac_dhw = 'PTAC_DHW'
    ptac = 'PTAC'
    pthp = 'PTHP'


class FurnaceEquipmentType(str, Enum):
    furnace = 'Furnace'
    furnace_electric = 'Furnace_Electric'


class VAV(_AllAirBase):
    """Variable Air Volume (VAV) HVAC system (aka. System 7 or 8).

    All rooms/zones are connected to a central air loop that is kept at a constant
    central temperature of 12.8C (55F). The central temperature is maintained by a
    cooling coil, which runs whenever the combination of return air and fresh outdoor
    air is greater than 12.8C, as well as a heating coil, which runs whenever
    the combination of return air and fresh outdoor air is less than 12.8C.

    Each air terminal for the connected rooms/zones contains its own reheat coil,
    which runs whenever the room is not in need of the cooling supplied by the 12.8C
    central air.

    The central cooling coil is always a chilled water coil, which is connected to a
    chilled water loop operating at 6.7C (44F). All heating coils are hot water coils
    except when Gas Coil equipment_type is used (in which case coils are gas)
    or when Parallel Fan-Powered (PFP) boxes equipment_type is used (in which case
    coils are electric resistance). Hot water temperature is 82C (180F) for
    boiler/district heating and 49C (120F) when ASHP is used.

    VAV systems are the traditional baseline system for commercial buildings
    taller than 5 stories or larger than 14,000 m2 (150,000 ft2) of floor area.
    """

    type: constr(regex='^VAV$') = 'VAV'

    equipment_type: VAVEquipmentType = Field(
        VAVEquipmentType.vav_chill_gbr,
        description='Text for the specific type of system equipment from the '
        'VAVEquipmentType enumeration.'
    )


class PVAV(_AllAirBase):
    """Packaged Variable Air Volume (PVAV) HVAC system (aka. System 5 or 6).

    All rooms/zones are connected to a central air loop that is kept at a constant
    central temperature of 12.8C (55F). The central temperature is maintained by a
    cooling coil, which runs whenever the combination of return air and fresh outdoor
    air is greater than 12.8C, as well as a heating coil, which runs whenever
    the combination of return air and fresh outdoor air is less than 12.8C.

    Each air terminal for the connected rooms/zones contains its own reheat coil,
    which runs whenever the room is not in need of the cooling supplied by the 12.8C
    central air.

    The central cooling coil is always a two-speed direct expansion (DX) coil.
    All heating coils are hot water coils except when Gas Coil equipment_type is
    used (in which case the central coil is gas and all reheat is electric)
    or when Parallel Fan-Powered (PFP) boxes equipment_type is used (in which case
    coils are electric resistance). Hot water temperature is 82C (180F) for
    boiler/district heating and 49C (120F) when ASHP is used.

    PVAV systems are the traditional baseline system for commercial buildings
    with than 4-5 stories or between 2,300 m2 and 14,000 m2 (25,000 ft2 and
    150,000 ft2) of floor area.
    """

    type: constr(regex='^PVAV$') = 'PVAV'

    equipment_type: PVAVEquipmentType = Field(
        PVAVEquipmentType.pvav_gbr,
        description='Text for the specific type of system equipment from the '
        'VAVEquipmentType enumeration.'
    )


class PSZ(_AllAirBase):
    """Packaged Single-Zone (PSZ) HVAC system (aka. System 3 or 4).

    Each room/zone receives its own air loop with its own single-speed direct expansion
    (DX) cooling coil, which will condition the supply air to a value in between
    12.8C (55F) and 50C (122F) depending on the heating/cooling needs of the room/zone.
    As long as a Baseboard equipment_type is NOT selected, heating will be supplied
    by a heating coil in the air loop. Otherwise, heating is accomplished with
    baseboards and the air loop only supplies cooling and ventilation air.
    Fans are constant volume.

    PSZ systems are the traditional baseline system for commercial buildings
    with less than 4 stories or less than 2,300 m2 (25,000 ft2) of floor area.
    They are also the default for all retail with less than 3 stories and all public
    assembly spaces.
    """

    type: constr(regex='^PSZ$') = 'PSZ'

    equipment_type: PSZEquipmentType = Field(
        PSZEquipmentType.psz_e_base,
        description='Text for the specific type of system equipment from the '
        'PVAVEquipmentType enumeration.'
    )


class PTAC(_TemplateSystem):
    """Packaged Terminal Air Conditioning (PTAC/HP) HVAC system. (aka. System 1 or 2).

    Each room/zone receives its own packaged unit that supplies heating, cooling
    and ventilation. Cooling is always done via a single-speed direct expansion (DX)
    cooling coil. Heating can be done via a heating coil in the unit or via an
    external baseboard. Fans are constant volume.

    PTAC/HP systems are the traditional baseline system for residential buildings.
    """

    type: constr(regex='^PTAC$') = 'PTAC'

    equipment_type: PTACEquipmentType = Field(
        PTACEquipmentType.ptac_e_base,
        description='Text for the specific type of system equipment from the '
        'PTACEquipmentType enumeration.'
    )


class ForcedAirFurnace(_TemplateSystem):
    """Forced Air Furnace HVAC system (aka. System 9 or 10).

    Forced air furnaces are intended only for spaces only requiring heating and
    ventilation. Each room/zone receives its own air loop with its own gas heating
    coil, which will supply air at a temperature up to 50C (122F) to meet the
    heating needs of the room/zone. Fans are constant volume.

    ForcedAirFurnace systems are the traditional baseline system for storage
    spaces that only require heating.
    """

    type: constr(regex='^ForcedAirFurnace$') = 'ForcedAirFurnace'

    equipment_type: FurnaceEquipmentType = Field(
        FurnaceEquipmentType.furnace,
        description='Text for the specific type of system equipment from the '
        'FurnaceEquipmentType enumeration.'
    )



"""Schemas for objects that generate electricity."""
from pydantic import Field, constr

from .._base import NoExtraBaseModel
from ._base import EnergyBaseModel
from enum import Enum


class ElectricLoadCenter(NoExtraBaseModel):

    type: constr(regex='^ElectricLoadCenter$') = 'ElectricLoadCenter'

    inverter_efficiency: float = Field(
        0.96,
        gt=0,
        le=1,
        description='A number between 0 and 1 for the load center inverter nominal '
        'rated DC-to-AC conversion efficiency. An inverter converts DC power, such '
        'as that output by photovoltaic panels, to AC power, such as that distributed '
        'by the electrical grid and is available from standard electrical outlets. '
        'Inverter efficiency is defined as the inverter rated AC power output divided '
        'by its rated DC power output.'
    )

    inverter_dc_to_ac_size_ratio: float = Field(
        1.1,
        gt=0,
        description='A positive number (typically greater than 1) for the ratio of '
        'the inverter DC rated size to its AC rated size. Typically, inverters are '
        'not sized to convert the full DC output under standard test conditions (STC) '
        'as such conditions rarely occur in reality and therefore unnecessarily add '
        'to the size/cost of the inverter. For a system with a high DC to AC size '
        'ratio, during times when the DC power output exceeds the inverter rated '
        'DC input size, the inverter limits the power output by increasing the '
        'DC operating voltage, which moves the arrays operating point down its '
        'current-voltage (I-V) curve. In EnergyPlus, this is accomplished by simply '
        'limiting the system output to the AC size as dictated by the ratio. The '
        'default value of 1.1 is reasonable for most systems. A typical range is '
        '1.1 to 1.25, although some large-scale systems have ratios of as high '
        'as 1.5. The optimal value depends on the system location, array '
        'orientation, and module cost.'
    )


class ModuleType(str, Enum):
    standard = 'Standard'
    premium = 'Premium'
    thin_film = 'ThinFilm'


class MountingType(str, Enum):
    fixed_open_rack = 'FixedOpenRack'
    fixed_roof_mounted = 'FixedRoofMounted'
    one_axis = 'OneAxis'
    one_axis_backtracking = 'OneAxisBacktracking'
    two_axis = 'TwoAxis'


class PVProperties(EnergyBaseModel):

    type: constr(regex='^PVProperties$') = 'PVProperties'

    rated_efficiency: float = Field(
        0.15,
        gt=0,
        lt=1,
        description='A number between 0 and 1 for the rated nameplate efficiency '
        'of the photovoltaic solar cells under standard test conditions (STC). '
        'Standard test conditions are 1,000 Watts per square meter solar '
        'irradiance, 25 degrees C cell temperature, and ASTM G173-03 standard '
        'spectrum. Nameplate efficiencies reported by manufacturers are typically '
        'under STC. Standard poly- or mono-crystalline silicon modules tend to have '
        'rated efficiencies in the range of 14-17%. Premium high efficiency '
        'mono-crystalline silicon modules with anti-reflective coatings can have '
        'efficiencies in the range of 18-20%. Thin film photovoltaic modules '
        'typically have efficiencies of 11% or less. (Default: 0.15 for standard '
        'silicon solar cells).'
    )

    active_area_fraction: float = Field(
        0.9,
        gt=0,
        le=1,
        description='The fraction of the parent Shade geometry that is '
        'covered in active solar cells. This fraction includes the difference '
        'between the PV panel (aka. PV module) area and the active cells within '
        'the panel as well as any losses for how the (typically rectangular) panels '
        'can be arranged on the Shade geometry. When the parent Shade geometry '
        'represents just the solar panels, this fraction is typically around 0.9 '
        'given that the framing elements of the panel reduce the overall '
        'active area. (Default: 0.9, assuming parent Shade geometry represents '
        'only the PV panel geometry).'
    )

    module_type: ModuleType = Field(
        default=ModuleType.standard,
        description='Text to indicate the type of solar module. This is used to '
        'determine the temperature coefficients used in the simulation of the '
        'photovoltaic modules. When the rated_efficiency is between 12-18%, the '
        'Standard type is typically most appropriate. When the rated_efficiency is '
        'greater than 18%, the Premium type is likely more appropriate. When '
        'the rated_efficiency is less than 12%, this likely refers to a case where '
        'the ThinFilm module type is most appropriate.'
    )

    mounting_type: MountingType = Field(
        default=MountingType.fixed_open_rack,
        description='Text to indicate the type of mounting and/or tracking used '
        'for the photovoltaic array. Note that the OneAxis options have an axis '
        'of rotation that is determined by the azimuth of the parent Shade '
        'geometry. Also note that, in the case of one or two axis tracking, '
        'shadows on the (static) parent Shade geometry still reduce the '
        'electrical output, enabling the simulation to account for large '
        'context geometry casting shadows on the array. However, the effects '
        'of smaller detailed shading may be improperly accounted for and self '
        'shading of the dynamic panel geometry is only accounted for via the '
        'tracking_ground_coverage_ratio property on this object. FixedOpenRack '
        'refers to ground or roof mounting where the air flows freely. FixedRoofMounted '
        'refers to mounting flush with the roof with limited air flow. OneAxis '
        'refers to a fixed tilt and azimuth, which define an axis of rotation. '
        'OneAxisBacktracking is the same as OneAxis but with controls to reduce '
        'self-shade at low sun angles. TwoAxis refers to a dynamic tilt and '
        'azimuth that track the sun.'
    )

    system_loss_fraction: float = Field(
        0.14,
        ge=0,
        le=1,
        description='A number between 0 and 1 for the fraction of the electricity '
        'output lost due to factors other than EPW weather conditions, '
        'panel efficiency/type, active area, mounting, and inverter conversion from '
        'DC to AC. Factors that should be accounted for in this input include '
        'soiling, snow, wiring losses, electrical connection losses, manufacturer '
        'defects/tolerances/mismatch in cell characteristics, losses from power '
        'grid availability, and losses due to age or light-induced degradation. '
        'Losses from these factors tend to be between 10-20% but can vary widely '
        'depending on the installation, maintenance and the grid to which the '
        'panels are connected..'
    )



"""Base class for HVAC systems following a standards template."""
from pydantic import Field
from enum import Enum

from .._base import IDdEnergyBaseModel


class Vintages(str, Enum):
    ashrae_2019 = 'ASHRAE_2019'
    ashrae_2016 = 'ASHRAE_2016'
    ashrae_2013 = 'ASHRAE_2013'
    ashrae_2010 = 'ASHRAE_2010'
    ashrae_2007 = 'ASHRAE_2007'
    ashrae_2004 = 'ASHRAE_2004'
    doe_ref_1980_2004 = 'DOE_Ref_1980_2004'
    doe_ref_pre_1980 = 'DOE_Ref_Pre_1980'


class _TemplateSystem(IDdEnergyBaseModel):
    """Base class for HVAC systems following a standards template."""

    vintage: Vintages = Field(
        Vintages.ashrae_2019,
        description='Text for the vintage of the template system. This will be used '
        'to set efficiencies for various pieces of equipment within the system. '
        'Further information about these defaults can be found in the version of '
        'ASHRAE 90.1 corresponding to the selected vintage. Read-only versions '
        'of the standard can be found at: https://www.ashrae.org/technical-resources/'
        'standards-and-guidelines/read-only-versions-of-ashrae-standards'
    )


class RadiantFaceTypes(str, Enum):
    floor = 'Floor'
    ceiling = 'Ceiling'
    floor_with_carpet = 'FloorWithCarpet'
    ceiling_metal_panel = 'CeilingMetalPanel'
    floor_with_hardwood = 'FloorWithHardwood'



from pydantic import Field, constr
from typing import List

from .._base import NoExtraBaseModel


class DaylightingControl(NoExtraBaseModel):

    type: constr(regex='^DaylightingControl$') = 'DaylightingControl'

    sensor_position: List[float] = Field(
        ...,
        description='A point as 3 (x, y, z) values for the position of the daylight '
        'sensor within the parent Room. This point should lie within the Room '
        'volume in order for the results to be meaningful.',
        min_items=3,
        max_items=3
    )

    illuminance_setpoint: float = Field(
        300,
        gt=0,
        description='A number for the illuminance setpoint in lux beyond '
        'which electric lights are dimmed if there is sufficient daylight.'
    )

    control_fraction: float = Field(
        1,
        ge=0,
        le=1,
        description='A number between 0 and 1 that represents the fraction of '
        'the Room lights that are dimmed when the illuminance at the sensor '
        'position is at the specified illuminance. 1 indicates that all lights are '
        'dim-able while 0 indicates that no lights are dim-able. Deeper rooms '
        'should have lower control fractions to account for the face that the '
        'lights in the back of the space do not dim in response to suitable '
        'daylight at the front of the room.'
    )

    min_power_input: float = Field(
        0.3,
        ge=0,
        le=1,
        description='A number between 0 and 1 for the the lowest power the '
        'lighting system can dim down to, expressed as a fraction of maximum '
        'input power.'
    )

    min_light_output: float = Field(
        0.2,
        ge=0,
        le=1,
        description='A number between 0 and 1 the lowest lighting output the '
        'lighting system can dim down to, expressed as a fraction of maximum '
        'light output.'
    )

    off_at_minimum: bool = Field(
        default=False,
        description='Boolean to note whether lights should switch off completely '
        'when they get to the minimum power input.'
    )



"""Construction Schema"""
from pydantic import Field, constr
from typing import List, Union
from enum import Enum

from ._base import IDdEnergyBaseModel
from .material import EnergyMaterial, EnergyMaterialNoMass, EnergyMaterialVegetation, \
    EnergyWindowMaterialGas, EnergyWindowMaterialGasCustom, \
    EnergyWindowMaterialGasMixture, EnergyWindowMaterialSimpleGlazSys, \
    EnergyWindowMaterialGlazing, EnergyWindowFrame, \
    EnergyWindowMaterialShade, EnergyWindowMaterialBlind
from .schedule import ScheduleRuleset, ScheduleFixedInterval


class OpaqueConstructionAbridged(IDdEnergyBaseModel):
    """Construction for opaque objects (Face, Shade, Door)."""

    type: constr(regex='^OpaqueConstructionAbridged$') = 'OpaqueConstructionAbridged'

    materials: List[constr(min_length=1, max_length=100)] = Field(
        ...,
        description='List of strings for opaque material identifiers. The order '
        'of the materials is from exterior to interior.',
        min_items=1,
        max_items=10
    )


class OpaqueConstruction(OpaqueConstructionAbridged):
    """Construction for opaque objects (Face, Shade, Door)."""

    type: constr(regex='^OpaqueConstruction$') = 'OpaqueConstruction'

    materials: List[
        Union[EnergyMaterial, EnergyMaterialNoMass, EnergyMaterialVegetation]
    ] = Field(
        ...,
        description='List of opaque material definitions. The order '
        'of the materials is from exterior to interior.',
        min_items=1,
        max_items=10
    )


class WindowConstructionAbridged(IDdEnergyBaseModel):
    """Construction for window objects (Aperture, Door)."""

    type: constr(regex='^WindowConstructionAbridged$') = 'WindowConstructionAbridged'

    materials: List[constr(min_length=1, max_length=100)] = Field(
        ...,
        description='List of strings for glazing or gas material identifiers. The '
        'order of the materials is from exterior to interior. If a SimpleGlazSys '
        'material is used, it must be the only material in the construction. '
        'For multi-layered constructions, adjacent glass layers must be separated '
        'by one and only one gas layer.',
        min_items=1,
        max_items=8
    )

    frame: str = Field(
        None,
        min_length=1,
        max_length=100,
        description='An optional identifier for a frame material that '
        'surrounds the window construction.'
    )


class WindowConstruction(WindowConstructionAbridged):
    """Construction for window objects (Aperture, Door)."""

    type: constr(regex='^WindowConstruction$') = 'WindowConstruction'

    materials: List[
        Union[
            EnergyWindowMaterialSimpleGlazSys, EnergyWindowMaterialGlazing,
            EnergyWindowMaterialGas, EnergyWindowMaterialGasCustom,
            EnergyWindowMaterialGasMixture
        ]
    ] = Field(
        ...,
        description='List of glazing and gas material definitions. The order '
        'of the materials is from exterior to interior. If a SimpleGlazSys '
        'material is used, it must be the only material in the construction. '
        'For multi-layered constructions, adjacent glass layers must be separated '
        'by one and only one gas layer.',
        min_items=1,
        max_items=8
    )

    frame: EnergyWindowFrame = Field(
        None,
        description='An optional window frame material for the frame that '
        'surrounds the window construction.'
    )


class ShadeLocation(str, Enum):
    """Choices for where a shade material is located in a window assembly."""
    interior = 'Interior'
    between = 'Between'
    exterior = 'Exterior'


class ControlType(str, Enum):
    """Choices for how the shading device is controlled."""
    always_on = 'AlwaysOn'
    on_if_high_solar_on_window = 'OnIfHighSolarOnWindow'
    on_if_high_horizontal_solar = 'OnIfHighHorizontalSolar'
    on_if_high_outdoor_air_temperature = 'OnIfHighOutdoorAirTemperature'
    on_if_high_zone_air_temperature = 'OnIfHighZoneAirTemperature'
    on_if_high_zone_cooling = 'OnIfHighZoneCooling'
    on_night_if_low_outdoor_temp_and_off_day = 'OnNightIfLowOutdoorTempAndOffDay'
    on_night_if_low_inside_temp_and_off_day = 'OnNightIfLowInsideTempAndOffDay'
    on_night_if_heating_and_off_day = 'OnNightIfHeatingAndOffDay'


class WindowConstructionShadeAbridged(IDdEnergyBaseModel):
    """Construction for window objects with an included shade layer."""

    type: constr(regex='^WindowConstructionShadeAbridged$') = \
        'WindowConstructionShadeAbridged'

    window_construction: WindowConstructionAbridged = Field(
        ...,
        description='A WindowConstructionAbridged object that serves as the '
        '"switched off" version of the construction (aka. the "bare construction"). '
        'The shade_material and shade_location will be used to modify this '
        'starting construction.'
    )

    shade_material: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='Identifier of a An EnergyWindowMaterialShade or an '
        'EnergyWindowMaterialBlind that serves as the shading layer for this '
        'construction. This can also be an EnergyWindowMaterialGlazing, which '
        'will indicate that the WindowConstruction has a dynamically-controlled '
        'glass pane like an electrochromic window assembly.'
    )

    shade_location: ShadeLocation = Field(
        ShadeLocation.interior,
        description='Text to indicate where in the window assembly the shade_material '
        'is located.  Note that the WindowConstruction must have at least one gas '
        'gap to use the "Between" option. Also note that, for a WindowConstruction '
        'with more than one gas gap, the "Between" option defaults to using the '
        'inner gap as this is the only option that EnergyPlus supports.'
    )

    control_type: ControlType = Field(
        ControlType.always_on,
        description='Text to indicate how the shading device is controlled, which '
        'determines when the shading is “on” or “off.”'
    )

    setpoint: float = Field(
        None,
        description='A number that corresponds to the specified control_type. '
        'This can be a value in (W/m2), (C) or (W) depending upon the control type.'
        'Note that this value cannot be None for any control type except "AlwaysOn."'
    )

    schedule: str = Field(
        None,
        min_length=1,
        max_length=100,
        description='An optional schedule identifier to be applied on top of the '
        'control_type. If None, the control_type will govern all behavior of '
        'the construction.'
    )


class WindowConstructionShade(WindowConstructionShadeAbridged):
    """Construction for window objects (Aperture, Door)."""

    type: constr(regex='^WindowConstructionShade$') = 'WindowConstructionShade'

    window_construction: WindowConstruction = Field(
        ...,
        description='A WindowConstruction object that serves as the "switched off" '
        'version of the construction (aka. the "bare construction"). '
        'The shade_material and shade_location will be used to modify this '
        'starting construction.'
    )

    shade_material: Union[
        EnergyWindowMaterialShade, EnergyWindowMaterialBlind,
        EnergyWindowMaterialGlazing
    ] = Field(
        ...,
        description='Identifier of a An EnergyWindowMaterialShade or an '
        'EnergyWindowMaterialBlind that serves as the shading layer for this '
        'construction. This can also be an EnergyWindowMaterialGlazing, which '
        'will indicate that the WindowConstruction has a dynamically-controlled '
        'glass pane like an electrochromic window assembly.'
    )

    schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        None,
        description='An optional ScheduleRuleset or ScheduleFixedInterval to be '
        'applied on top of the control_type. If None, the control_type will govern '
        'all behavior of the construction.'
    )


class WindowConstructionDynamicAbridged(IDdEnergyBaseModel):
    """Construction for window objects with an included shade layer."""

    type: constr(regex='^WindowConstructionDynamicAbridged$') = \
        'WindowConstructionDynamicAbridged'

    constructions: List[WindowConstructionAbridged] = Field(
        ...,
        description='A list of WindowConstructionAbridged objects that define the '
        'various states that the dynamic window can assume.'
    )

    schedule: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='An identifier for a control schedule that dictates which '
        'constructions are active at given times throughout the simulation. The values '
        'of the schedule should be integers and range from 0 to one less then the '
        'number of constructions. Zero indicates that the first construction is active, '
        'one indicates that the second on is active, etc. The schedule type limits '
        'of this schedule should be "Control Level." If building custom schedule '
        'type limits that describe a particular range of states, the type limits '
        'should be "Discrete" and the unit type should be "Mode," "Control," or '
        'some other fractional unit.'
    )


class WindowConstructionDynamic(IDdEnergyBaseModel):
    """Construction for window objects with an included shade layer."""

    type: constr(regex='^WindowConstructionDynamic$') = \
        'WindowConstructionDynamic'

    constructions: List[WindowConstruction] = Field(
        ...,
        description='A list of WindowConstruction objects that define the '
        'various states that the dynamic window can assume.'
    )

    schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        ...,
        description='A control schedule that dictates which constructions '
        'are active at given times throughout the simulation. The values of the '
        'schedule should be integers and range from 0 to one less then the number '
        'of constructions. Zero indicates that the first construction is active, '
        'one indicates that the second on is active, etc. The schedule type limits '
        'of this schedule should be "Control Level." If building custom schedule '
        'type limits that describe a particular range of states, the type limits '
        'should be "Discrete" and the unit type should be "Mode," "Control," or '
        'some other fractional unit.'
    )


class ShadeConstruction(IDdEnergyBaseModel):
    """Construction for Shade objects."""

    type: constr(regex='^ShadeConstruction$') = 'ShadeConstruction'

    solar_reflectance: float = Field(
        0.2,
        ge=0,
        le=1,
        description='A number for the solar reflectance of the construction.'
    )

    visible_reflectance: float = Field(
        0.2,
        ge=0,
        le=1,
        description='A number for the visible reflectance of the construction.'
    )

    is_specular: bool = Field(
        default=False,
        description='Boolean to note whether the reflection off the shade is diffuse '
        '(False) or specular (True). Set to True if the construction is '
        'representing a glass facade or a mirror material.'
    )


class AirBoundaryConstructionAbridged(IDdEnergyBaseModel):
    """Construction for Air Boundary objects."""

    type: constr(regex='^AirBoundaryConstructionAbridged$') = \
        'AirBoundaryConstructionAbridged'

    air_mixing_per_area: float = Field(
        0.1,
        ge=0,
        description='A positive number for the amount of air mixing between Rooms '
        'across the air boundary surface [m3/s-m2]. Default: 0.1 corresponds '
        'to average indoor air speeds of 0.1 m/s (roughly 20 fpm), which is '
        'typical of what would be induced by a HVAC system.'
    )

    air_mixing_schedule: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of a fractional schedule for the air mixing schedule '
        'across the construction. If unspecified, an Always On schedule will be assumed.'
    )


class AirBoundaryConstruction(AirBoundaryConstructionAbridged):
    """Construction for Air Boundary objects."""

    type: constr(regex='^AirBoundaryConstruction$') = 'AirBoundaryConstruction'

    air_mixing_schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        default=None,
        description='A fractional schedule as a ScheduleRuleset or '
        'ScheduleFixedInterval for the air mixing schedule across '
        'the construction. If unspecified, an Always On schedule will be assumed.'
    )



"""Boundary condition schemas."""
from pydantic import Field, constr
from typing import List, Union

from ._base import NoExtraBaseModel
from .altnumber import Autocalculate


class Outdoors(NoExtraBaseModel):

    type: constr(regex='^Outdoors$') = 'Outdoors'

    sun_exposure: bool = Field(
        True,
        description='A boolean noting whether the boundary is exposed to sun.'
    )

    wind_exposure: bool = Field(
        True,
        description='A boolean noting whether the boundary is exposed to wind.'
    )

    view_factor: Union[Autocalculate, float] = Field(
        Autocalculate(),
        ge=0,
        le=1,
        description='A number for the view factor to the ground. This can also be '
        'an Autocalculate object to have the view factor automatically calculated.'
    )


class Surface(NoExtraBaseModel):

    type: constr(regex='^Surface$') = 'Surface'

    boundary_condition_objects: List[str] = Field(
        ...,
        min_items=2,
        max_items=3,
        description='A list of up to 3 object identifiers that are adjacent to this one.'
        ' The first object is always the one that is immediately adjacent and is of '
        'the same object type (Face, Aperture, Door). When this boundary condition '
        'is applied to a Face, the second object in the tuple will be the parent '
        'Room of the adjacent object. When the boundary condition is applied to a '
        'sub-face (Door or Aperture), the second object will be the parent Face '
        'of the adjacent sub-face and the third object will be the parent Room '
        'of the adjacent sub-face.'
    )


class Ground(NoExtraBaseModel):

    type: constr(regex='^Ground$') = 'Ground'


class Adiabatic(NoExtraBaseModel):

    type: constr(regex='^Adiabatic$') = 'Adiabatic'


class OtherSideTemperature(NoExtraBaseModel):

    type: constr(regex='^OtherSideTemperature$') = 'OtherSideTemperature'

    heat_transfer_coefficient: float = Field(
        0,
        ge=0,
        description='A value in W/m2-K to indicate the combined convective/radiative '
        'film coefficient. If equal to 0, then the specified temperature above is '
        'equal to the exterior surface temperature. Otherwise, the temperature above '
        'is considered the outside air temperature and this coefficient is used to '
        'determine the difference between this outside air temperature and the '
        'exterior surface temperature.'
    )

    temperature: Union[Autocalculate, float] = Field(
        Autocalculate(),
        description='A temperature value in Celsius to note the temperature on the '
        'other side of the object. This input can also be an Autocalculate object '
        'to signify that the temperature is equal to the outdoor air temperature.'
    )



"""Command Line Interface (CLI) entry point for honeybee schema."""

try:
    import click
except ImportError:
    raise ImportError(
        'click module is not installed. Try `pip install honeybee-schema[cli]` command.'
    )

import sys
import logging
import json
import importlib
from inspect import getmembers, isfunction
import pkgutil

from honeybee_schema import updater


@click.group()
@click.version_option()
def main():
    pass


_logger = logging.getLogger(__name__)


@main.command('update-model')
@click.argument('model-json', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.option('--version', '-v', help='Text to indicate the version to which the model '
              'JSON will be updated (eg. 1.41.2). Versions must always consist of '
              'three integers separated b periods. If None, the Model JSON will '
              'be updated to the last release that included a breaking change.',
              type=str, default=None)
@click.option('--output-file', help='Optional file to output the JSON string of '
              'the config object. By default, it will be printed out to stdout',
              type=click.File('w'), default='-', show_default=True)
def update_model(model_json, version, output_file):
    """Update a Honeybee Model JSON to a newer version of honeybee-schema.

    \b
    Args:
        model_json: Full path to a Model JSON file.
    """
    try:
        # get the version to which the model will be updated
        up_version = (999, 999, 999) if version is None else \
            tuple(int(v) for v in version.split('.'))

        # load the model dictionary from the json file and get the version of it
        with open(model_json) as json_file:
            model_dict = json.load(json_file)
        assert 'version' in model_dict, 'No version was found in the input model ' \
            'JSON. Update process cannot be run.'
        print(f'Input model version: {model_dict["version"]}', file=sys.stderr)
        model_version = tuple(int(v) for v in model_dict['version'].split('.'))

        if model_version >= up_version:
            # no point for updating. Normally we should assert here but by writing
            # the same file to a new file we make this command more flexible for cases
            # that we don't know the version for the input model and we want to update
            # it just in case.
            mv = '.'.join(map(str, model_version))
            uv = '.'.join(map(str, up_version))
            print(
                f'The input file has a higher version ({mv}) than the target version'
                f' ({uv}).', file=sys.stderr
            )
            output_file.write(json.dumps(model_dict))
            # let's consider exporting the same file as success
            sys.exit(0)

        # get functions with a higher version than model_version
        updaters = []

        for sub_module in pkgutil.walk_packages(updater.__path__):
            if not sub_module.name.startswith('version_'):
                continue
            module = importlib.import_module(f'honeybee_schema.updater.{sub_module.name}')
            for name, func in getmembers(module, isfunction):
                if not name.startswith('version_'):
                    continue
                func_version = tuple(int(v) for v in name.split('_')[1:])
                if model_version > func_version or func_version > up_version:
                    continue
                updaters.append((func_version, func))

        # sort updaters based on version
        updaters = sorted(updaters, key=lambda x: x[0])

        # update the dictionary
        for func_version, up_func in updaters:
            print(
                f'Updating to version {".".join(map(str, func_version))}',
                file=sys.stderr
            )
            model_dict = up_func(model_dict)

        # update the model dictionary version and write it to log file
        if version:
            model_dict['version'] = version
        elif updaters:
            model_dict['version'] = '.'.join((str(i) for i in updaters[-1][0]))
        output_file.write(json.dumps(model_dict))
    except Exception as e:
        _logger.exception('Failed to update Honeybee Model JSON.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()



"""Material Schema"""
from pydantic import Field, validator, root_validator, constr, confloat
from typing import Union, List
from enum import Enum

from ._base import IDdEnergyBaseModel
from ..altnumber import Autocalculate


class Roughness(str, Enum):
    """Relative roughness of a particular material layer."""
    very_rough = 'VeryRough'
    rough = 'Rough'
    medium_rough = 'MediumRough'
    medium_smooth = 'MediumSmooth'
    smooth = 'Smooth'
    very_smooth = 'VerySmooth'


class MoistureDiffusionModel(str, Enum):
    """Acceptable values for the moisture diffusion model for vegetation."""
    simple = 'Simple'
    advanced = 'Advanced'


class EnergyMaterialNoMass(IDdEnergyBaseModel):
    """No mass opaque material representing a layer within an opaque construction.

    Used when only the thermal resistance (R value) of the material is known.
    """

    type: constr(regex='^EnergyMaterialNoMass$') = 'EnergyMaterialNoMass'

    r_value: float = Field(
        ...,
        ge=0.001,
        description='The thermal resistance (R-value) of the material layer [m2-K/W].'
    )

    roughness: Roughness = Roughness.medium_rough

    thermal_absorptance: float = Field(
        0.9,
        gt=0,
        le=0.99999,
        description='Fraction of incident long wavelength radiation that is absorbed by'
        ' the material. Default: 0.9.'
    )

    solar_absorptance: float = Field(
        0.7,
        ge=0,
        le=1,
        description='Fraction of incident solar radiation absorbed by the material.'
        ' Default: 0.7.'
    )

    visible_absorptance: float = Field(
        0.7,
        ge=0,
        le=1,
        description='Fraction of incident visible wavelength radiation absorbed by the'
        ' material. Default: 0.7.'
    )


class EnergyMaterial(IDdEnergyBaseModel):
    """Opaque material representing a layer within an opaque construction."""

    type: constr(regex='^EnergyMaterial$') = 'EnergyMaterial'

    roughness: Roughness = Roughness.medium_rough

    thickness: float = Field(
        ...,
        gt=0,
        le=3,
        description='Thickness of the material layer in meters.'
    )

    conductivity: float = Field(
        ...,
        gt=0,
        description='Thermal conductivity of the material layer in W/m-K.',
    )

    density: float = Field(
        ...,
        gt=0,
        description='Density of the material layer in kg/m3.'
    )

    specific_heat: float = Field(
        ...,
        ge=100,
        description='Specific heat of the material layer in J/kg-K.'
    )

    thermal_absorptance: float = Field(
        0.9,
        gt=0,
        le=0.99999,
        description='Fraction of incident long wavelength radiation that is absorbed by'
        ' the material. Default: 0.9.'
    )

    solar_absorptance: float = Field(
        0.7,
        ge=0,
        le=1,
        description='Fraction of incident solar radiation absorbed by the material.'
        ' Default: 0.7.'
    )

    visible_absorptance: float = Field(
        0.7,
        ge=0,
        le=1,
        description='Fraction of incident visible wavelength radiation absorbed by the'
        ' material. Default: 0.7.'
    )


class EnergyMaterialVegetation(IDdEnergyBaseModel):
    """Material representing vegetation on the exterior of an opaque construction."""

    type: constr(regex='^EnergyMaterialVegetation$') = 'EnergyMaterialVegetation'

    roughness: Roughness = Roughness.medium_rough

    thickness: float = Field(
        0.1,
        gt=0,
        le=3,
        description='Thickness of the soil layer in meters.'
    )

    conductivity: float = Field(
        0.35,
        gt=0,
        description='Thermal conductivity of the dry soil in W/m-K.',
    )

    density: float = Field(
        1100,
        gt=0,
        description='Density of the dry soil in kg/m3.'
    )

    specific_heat: float = Field(
        1200,
        ge=100,
        description='Specific heat of the dry soil in J/kg-K.'
    )

    soil_thermal_absorptance: float = Field(
        0.9,
        gt=0,
        le=0.99999,
        description='Fraction of incident long wavelength radiation that is absorbed by'
        ' the soil. Default: 0.9.'
    )

    soil_solar_absorptance: float = Field(
        0.7,
        ge=0,
        le=1,
        description='Fraction of incident solar radiation absorbed by the soil.'
        ' Default: 0.7.'
    )

    soil_visible_absorptance: float = Field(
        0.7,
        ge=0,
        le=1,
        description='Fraction of incident visible wavelength radiation absorbed by the'
        ' material. Default: 0.7.'
    )

    plant_height: float = Field(
        0.2,
        ge=0.005,
        le=1.0,
        description='The height of plants in the vegetation in meters.'
    )

    leaf_area_index: float = Field(
        1.0,
        ge=0.001,
        le=5.0,
        description='The projected leaf area per unit area of soil surface '
        '(aka. Leaf Area Index or LAI). Note that the fraction of vegetation '
        'cover is calculated directly from LAI using an empirical relation.'
    )

    leaf_reflectivity: float = Field(
        0.22,
        ge=0.005,
        le=0.5,
        description='The fraction of incident solar radiation that is reflected '
        'by the leaf surfaces. Solar radiation includes the visible spectrum as '
        'well as infrared and ultraviolet wavelengths. Typical values are 0.18 to 0.25.'
    )

    leaf_emissivity: float = Field(
        0.95,
        ge=0.8,
        le=1.0,
        description='The ratio of thermal radiation emitted from leaf surfaces to '
        'that emitted by an ideal black body at the same temperature.'
    )

    min_stomatal_resist: float = Field(
        180,
        ge=50,
        le=300,
        description='The resistance of the plants to moisture transport [s/m]. '
        'Plants with low values of stomatal resistance will result in higher '
        'evapotranspiration rates than plants with high resistance.'
    )

    sat_vol_moist_cont: float = Field(
        0.3,
        ge=0.1,
        le=0.5,
        description='The saturation moisture content of the soil by volume.'
    )

    residual_vol_moist_cont: float = Field(
        0.01,
        ge=0.01,
        le=0.1,
        description='The residual moisture content of the soil by volume.'
    )

    init_vol_moist_cont: float = Field(
        0.01,
        ge=0.05,
        le=0.5,
        description='The initial moisture content of the soil by volume.'
    )

    moist_diff_model: MoistureDiffusionModel = MoistureDiffusionModel.simple


class EnergyWindowMaterialSimpleGlazSys(IDdEnergyBaseModel):
    """Describe an entire glazing system rather than individual layers.

    Used when only very limited information is available on the glazing layers or when
    specific performance levels are being targeted.
    """

    type: constr(regex='^EnergyWindowMaterialSimpleGlazSys$') = \
        'EnergyWindowMaterialSimpleGlazSys'

    u_factor: float = Field(
        ...,
        gt=0,
        le=12,
        description='The overall heat transfer coefficient for window system in W/m2-K. '
        'Note that constructions with U-values above 5.8 should not be assigned to '
        'skylights as this implies the resistance of the window is negative when '
        'air films are subtracted.'
    )

    shgc: float = Field(
        ...,
        gt=0,
        lt=1,
        description='Unit-less quantity for the Solar Heat Gain Coefficient (solar '
        'transmittance + conduction) at normal incidence and vertical orientation.'
    )

    vt: float = Field(
        0.54,
        gt=0,
        lt=1,
        description='The fraction of visible light falling on the window that makes it'
        ' through the glass at normal incidence.'
    )


class EnergyWindowMaterialGlazing(IDdEnergyBaseModel):
    """Describe a single glass pane corresponding to a layer in a window construction."""

    type: constr(regex='^EnergyWindowMaterialGlazing$') = 'EnergyWindowMaterialGlazing'

    thickness: float = Field(
        0.003,
        gt=0,
        description='The surface-to-surface thickness of the glass in meters. Default: '
        ' 0.003.'
    )

    solar_transmittance: float = Field(
        0.85,
        ge=0,
        le=1,
        description='Transmittance of solar radiation through the glass at normal'
        ' incidence. Default: 0.85 for clear glass.'
    )

    solar_reflectance: float = Field(
        0.075,
        ge=0,
        le=1,
        description='Reflectance of solar radiation off of the front side of the glass'
        ' at normal incidence, averaged over the solar spectrum. Default: 0.075'
        ' for clear glass.'
    )

    solar_reflectance_back: Union[Autocalculate, float] = Field(
        default=Autocalculate(),
        ge=0,
        le=1,
        description='Reflectance of solar radiation off of the back side of the glass at'
        ' normal incidence, averaged over the solar spectrum.'
    )

    visible_transmittance: float = Field(
        0.9,
        ge=0,
        le=1,
        description='Transmittance of visible light through the glass at '
        'normal incidence. Default: 0.9 for clear glass.'
    )

    visible_reflectance: float = Field(
        0.075,
        ge=0,
        le=1,
        description='Reflectance of visible light off of the front side of the glass at'
        ' normal incidence. Default: 0.075 for clear glass.'
    )

    visible_reflectance_back: Union[Autocalculate, float] = Field(
        default=Autocalculate(),
        ge=0,
        le=1,
        description='Reflectance of visible light off of the back side of the glass at'
        ' normal incidence averaged over the solar spectrum and weighted by the response'
        ' of the human eye.'
    )

    infrared_transmittance: float = Field(
        0,
        ge=0,
        le=1,
        description='Long-wave transmittance at normal incidence.'
    )

    emissivity: float = Field(
        0.84,
        ge=0,
        le=1,
        description='Infrared hemispherical emissivity of the front (outward facing)'
        ' side of the glass.  Default: 0.84, which is typical for clear glass'
        ' without a low-e coating.'
    )

    emissivity_back: float = Field(
        0.84,
        ge=0,
        le=1,
        description='Infrared hemispherical emissivity of the back (inward facing) side'
        ' of the glass.  Default: 0.84, which is typical for clear glass without'
        ' a low-e coating.'
    )

    conductivity: float = Field(
        0.9,
        gt=0,
        description='Thermal conductivity of the glass in W/(m-K). Default: 0.9,'
        ' which is  typical for clear glass without a low-e coating.'
    )

    dirt_correction: float = Field(
        1,
        description='Factor that corrects for the presence of dirt on the glass. A'
        ' default value of 1 indicates the glass is clean.'
    )

    solar_diffusing: bool = Field(
        False,
        description='If False (default), the beam solar radiation incident on the '
        'glass is transmitted as beam radiation with no diffuse component.'
        'If True, the beam  solar radiation incident on the glass is transmitted as '
        'hemispherical diffuse radiation with no beam component.'
    )


class EnergyWindowFrame(IDdEnergyBaseModel):
    """Opaque material representing a layer within an opaque construction."""

    type: constr(regex='^EnergyWindowFrame$') = 'EnergyWindowFrame'

    width: float = Field(
        ...,
        gt=0,
        le=1,
        description='Number for the width of frame in plane of window [m]. '
        'The frame width is assumed to be the same on all sides of window..'
    )

    conductance: float = Field(
        ...,
        gt=0,
        description='Number for the thermal conductance of the frame material '
        'measured from inside to outside of the frame surface (no air films) '
        'and taking 2D conduction effects into account [W/m2-K].',
    )

    edge_to_center_ratio: float = Field(
        1,
        gt=0,
        le=4,
        description='Number between 0 and 4 for the ratio of the glass '
        'conductance near the frame (excluding air films) divided by the glass '
        'conductance at the center of the glazing (excluding air films). '
        'This is used only for multi-pane glazing constructions. This ratio should '
        'usually be greater than 1.0 since the spacer material that separates '
        'the glass panes is usually more conductive than the gap between panes. '
        'A value of 1 effectively indicates no spacer. Values should usually be '
        'obtained from the LBNL WINDOW program so that the unique characteristics '
        'of the window construction can be accounted for.'
    )

    outside_projection: float = Field(
        0,
        ge=0,
        le=0.5,
        description='Number for the distance that the frame projects outward from '
        'the outside face of the glazing [m]. This is used to calculate shadowing '
        'of frame onto glass, solar absorbed by the frame, IR emitted and '
        'absorbed by the frame, and convection from frame.'
    )

    inside_projection: float = Field(
        0,
        ge=0,
        le=0.5,
        description='Number for the distance that the frame projects inward from '
        'the inside face of the glazing [m]. This is used to calculate solar '
        'absorbed by the frame, IR emitted and absorbed by the frame, and '
        'convection from frame.'
    )

    thermal_absorptance: float = Field(
        0.9,
        gt=0,
        le=0.99999,
        description='Fraction of incident long wavelength radiation that is absorbed by'
        ' the frame material.'
    )

    solar_absorptance: float = Field(
        0.7,
        ge=0,
        le=1,
        description='Fraction of incident solar radiation absorbed by the '
        'frame material.'
    )

    visible_absorptance: float = Field(
        0.7,
        ge=0,
        le=1,
        description='Fraction of incident visible wavelength radiation absorbed by the'
        ' frame material.'
    )


class GasType (str, Enum):
    air = 'Air'
    argon = 'Argon'
    krypton = 'Krypton'
    xenon = 'Xenon'


class EnergyWindowMaterialGas(IDdEnergyBaseModel):
    """Create single layer of gas in a window construction.

    Can be combined with EnergyWindowMaterialGlazing to make multi-pane windows.
    """

    type: constr(regex='^EnergyWindowMaterialGas$') = 'EnergyWindowMaterialGas'

    thickness: float = Field(
        0.0125,
        gt=0,
        description='Thickness of the gas layer in meters. Default: 0.0125.'
    )

    gas_type: GasType = GasType.air


class EnergyWindowMaterialGasMixture(IDdEnergyBaseModel):
    """Create a mixture of two to four different gases to fill the panes of multiple
    pane windows."""

    type: constr(regex='^EnergyWindowMaterialGasMixture$') = \
        'EnergyWindowMaterialGasMixture'

    thickness: float = Field(
        0.0125,
        gt=0,
        description='The thickness of the gas mixture layer in meters.'
    )

    gas_types: List[GasType] = Field(
        ...,
        min_items=2,
        max_items=4,
        description='List of gases in the gas mixture.'
    )

    gas_fractions: List[confloat(gt=0, lt=1)] = Field(
        ...,
        min_items=2,
        max_items=4,
        description='A list of fractional numbers describing the volumetric fractions '
        'of gas types in the mixture. This list must align with the gas_types '
        'list and must sum to 1.'
    )

    @validator('gas_fractions')
    def check_sum(cls, v):
        """Check that fractions sum to 1."""
        assert abs(1 - sum(v)) < 0.001, 'gas_fractions must sum to 1.'
        return v

    @root_validator
    def check_types_fractions_match(cls, values):
        "Ensure the gas types and fractions match."
        gas_types = values.get('gas_types')
        gas_fractions = values.get('gas_fractions')
        assert len(gas_types) == len(gas_fractions), 'Length of gas_types must match ' \
            'length of gas_fractions. {} != {}'.format(
                len(gas_types), len(gas_fractions))
        return values


class EnergyWindowMaterialGasCustom(IDdEnergyBaseModel):
    """Create single layer of custom gas."""

    type: constr(regex='^EnergyWindowMaterialGasCustom$') = \
        'EnergyWindowMaterialGasCustom'

    thickness: float = Field(
        0.0125,
        gt=0,
        description='Thickness of the gas layer in meters. Default: 0.0125.'
    )

    conductivity_coeff_a: float = Field(
        ...,
        description='The A coefficient for gas conductivity in W/(m-K).'
    )

    conductivity_coeff_b: float = Field(
        0,
        description='The B coefficient for gas conductivity in W/(m-K2).'
    )

    conductivity_coeff_c: float = Field(
        0,
        description='The C coefficient for gas conductivity in W/(m-K3).'
    )

    viscosity_coeff_a: float = Field(
        ...,
        gt=0,
        description='The A coefficient for gas viscosity in kg/(m-s).'
    )

    viscosity_coeff_b: float = Field(
        0,
        description='The B coefficient for gas viscosity in kg/(m-s-K).'
    )

    viscosity_coeff_c: float = Field(
        0,
        description='The C coefficient for gas viscosity in kg/(m-s-K2).'
    )

    specific_heat_coeff_a: float = Field(
        ...,
        gt=0,
        description='The A coefficient for gas specific heat in J/(kg-K).'
    )

    specific_heat_coeff_b: float = Field(
        0,
        description='The B coefficient for gas specific heat in J/(kg-K2).'
    )

    specific_heat_coeff_c: float = Field(
        0,
        description='The C coefficient for gas specific heat in J/(kg-K3).'
    )

    specific_heat_ratio: float = Field(
        ...,
        gt=1,
        description='The specific heat ratio for gas.'
    )

    molecular_weight: float = Field(
        ...,
        ge=20,
        le=200,
        description='The molecular weight for gas in g/mol.'
    )


class EnergyWindowMaterialShade(IDdEnergyBaseModel):
    """This object specifies the properties of window shade materials."""

    type: constr(regex='^EnergyWindowMaterialShade$') = 'EnergyWindowMaterialShade'

    solar_transmittance: float = Field(
        0.4,
        ge=0,
        lt=1,
        description='The transmittance averaged over the solar spectrum. It is assumed'
        ' independent of incidence angle. Default: 0.4.'
    )

    solar_reflectance: float = Field(
        0.5,
        ge=0,
        lt=1,
        description='The reflectance averaged over the solar spectrum. It us assumed'
        ' same on both sides of shade and independent of incidence angle. Default'
        ' value is 0.5'
    )

    visible_transmittance: float = Field(
        0.4,
        ge=0,
        lt=1,
        description='The transmittance averaged over the solar spectrum and weighted by'
        ' the response of the human eye. It is assumed independent of incidence angle.'
        ' Default: 0.4.'
    )

    visible_reflectance: float = Field(
        0.4,
        ge=0,
        lt=1,
        description='The transmittance averaged over the solar spectrum and weighted by'
        ' the response of the human eye. It is assumed independent of incidence angle.'
        ' Default: 0.4'
    )

    emissivity: float = Field(
        0.9,
        gt=0,
        lt=1,
        description='The effective long-wave infrared hemispherical emissivity. It is'
        ' assumed same on both sides of shade. Default: 0.9.'
    )

    infrared_transmittance: float = Field(
        0,
        ge=0,
        lt=1,
        description='The effective long-wave transmittance. It is assumed independent'
        ' of incidence angle. Default: 0.'
    )

    thickness: float = Field(
        0.005,
        gt=0,
        description='The thickness of the shade material in meters. Default:'
        ' 0.005.'
    )

    conductivity: float = Field(
        0.1,
        gt=0,
        description='The conductivity of the shade material in W/(m-K). Default value'
        ' is 0.1.'
    )

    distance_to_glass: float = Field(
        0.05,
        ge=0.001,
        le=1,
        description='The distance from shade to adjacent glass in meters. Default value'
        ' is 0.05'
    )

    top_opening_multiplier: float = Field(
        0.5,
        ge=0,
        le=1,
        description='The effective area for air flow at the top of the shade, divided by'
        ' the horizontal area between glass and shade. Default: 0.5.'
    )

    bottom_opening_multiplier: float = Field(
        0.5,
        ge=0,
        le=1,
        description='The effective area for air flow at the bottom of the shade, divided'
        ' by the horizontal area between glass and shade. Default: 0.5.'
    )

    left_opening_multiplier: float = Field(
        0.5,
        ge=0,
        le=1,
        description='The effective area for air flow at the left side of the shade,'
        ' divided by the vertical area between glass and shade. Default: 0.5.'
    )

    right_opening_multiplier: float = Field(
        0.5,
        ge=0,
        le=1,
        description='The effective area for air flow at the right side of the shade,'
        ' divided by the vertical area between glass and shade. Default: 0.5.'
    )

    airflow_permeability: float = Field(
        0,
        ge=0,
        le=0.8,
        description='The fraction of the shade surface that is open to air flow.'
        ' If air cannot pass through the shade material, airflow_permeability = 0.'
        ' Default: 0.'
    )


class SlatOrientation (str, Enum):
    horizontal = 'Horizontal'
    vertical = 'Vertical'


class EnergyWindowMaterialBlind(IDdEnergyBaseModel):
    """Window blind material consisting of flat, equally-spaced slats."""

    type: constr(regex='^EnergyWindowMaterialBlind$') = 'EnergyWindowMaterialBlind'

    slat_orientation: SlatOrientation = SlatOrientation.horizontal

    slat_width: float = Field(
        0.025,
        gt=0,
        le=1,
        description='The width of slat measured from edge to edge in meters.'
    )

    slat_separation: float = Field(
        0.01875,
        gt=0,
        le=1,
        description='The distance between the front of a slat and the back of the'
        ' adjacent slat in meters.'
    )

    slat_thickness: float = Field(
        0.001,
        gt=0,
        le=0.1,
        description='The distance between the faces of a slat in meters. The default'
        ' value is 0.001.'
    )

    slat_angle: float = Field(
        45,
        ge=0,
        le=180,
        description='The angle (degrees) between the glazing outward normal and the slat'
        ' outward normal where the outward normal points away from the front face of the'
        ' slat (degrees). The default value is 45.'
    )

    slat_conductivity: float = Field(
        221,
        gt=0,
        description='The thermal conductivity of the slat in W/(m-K). Default:'
        ' 221.'
    )

    beam_solar_transmittance: float = Field(
        0,
        ge=0,
        lt=1,
        description='The beam solar transmittance of the slat, assumed to be independent'
        ' of angle of incidence on the slat. Any transmitted beam radiation is assumed'
        ' to be 100% diffuse (i.e., slats are translucent). The default value is 0.'
    )

    beam_solar_reflectance: float = Field(
        0.5,
        ge=0,
        lt=1,
        description='The beam solar reflectance of the front side of the slat, it is'
        ' assumed to be independent of the angle of incidence. Default: 0.5.'
    )

    beam_solar_reflectance_back: float = Field(
        0.5,
        ge=0,
        lt=1,
        description='The beam solar reflectance of the back side of the slat, it is'
        ' assumed to be independent of the angle of incidence. Default: 0.5.'
    )

    diffuse_solar_transmittance: float = Field(
        0,
        ge=0,
        lt=1,
        description='The slat transmittance for hemispherically diffuse solar radiation.'
        ' Default: 0.'
    )

    diffuse_solar_reflectance: float = Field(
        0.5,
        ge=0,
        lt=1,
        description='The front-side slat reflectance for hemispherically diffuse solar'
        ' radiation. Default: 0.5.'
    )

    diffuse_solar_reflectance_back: float = Field(
        0.5,
        ge=0,
        lt=1,
        description='The back-side slat reflectance for hemispherically diffuse solar'
        ' radiation. Default: 0.5.'
    )

    beam_visible_transmittance: float = Field(
        0,
        ge=0,
        lt=1,
        description='The beam visible transmittance of the slat, it is assumed to be'
        ' independent of the angle of incidence. Default: 0.'
    )

    beam_visible_reflectance: float = Field(
        0.5,
        ge=0,
        lt=1,
        description='The beam visible reflectance on the front side of the slat, it is'
        ' assumed to be independent of the angle of incidence. Default: 0.5.'
    )

    beam_visible_reflectance_back: float = Field(
        0.5,
        ge=0,
        lt=1,
        description='The beam visible reflectance on the back side of the slat, it is'
        ' assumed to be independent of the angle of incidence. Default: 0.5.'
    )

    diffuse_visible_transmittance: float = Field(
        0,
        ge=0,
        lt=1,
        description='The slat transmittance for hemispherically diffuse visible'
        ' radiation. This value should equal “Slat Beam Visible Transmittance.”'
    )

    diffuse_visible_reflectance: float = Field(
        0.5,
        ge=0,
        lt=1,
        description='The front-side slat reflectance for hemispherically diffuse visible'
        ' radiation. This value should equal “Front Side Slat Beam Visible Reflectance.”'
        ' Default: 0.5.'
    )

    diffuse_visible_reflectance_back: float = Field(
        0.5,
        ge=0,
        lt=1,
        description='The back-side slat reflectance for hemispherically diffuse visible'
        ' radiation. This value should equal “Back Side Slat Beam Visible Reflectance.'
        ' Default: 0.5.”'
    )

    infrared_transmittance: float = Field(
        0,
        ge=0,
        lt=1,
        description='The slat infrared hemispherical transmittance. It is zero for '
        'solid metallic, wooden or glass slats, but may be non-zero in some cases '
        'such as for thin plastic slats. The default value is 0.'
    )

    emissivity: float = Field(
        0.9,
        ge=0,
        lt=1,
        description='Front side hemispherical emissivity of the slat. Default is 0.9 for'
        ' most materials. The default value is 0.9.'
    )

    emissivity_back: float = Field(
        0.9,
        ge=0,
        lt=1,
        description='Back side hemispherical emissivity of the slat. Default is 0.9 for'
        ' most materials. The default value is 0.9.'
    )

    distance_to_glass: float = Field(
        0.05,
        ge=0.01,
        le=1,
        description='The distance from the mid-plane of the blind to the adjacent glass'
        ' in meters. The default value is 0.05.'
    )

    top_opening_multiplier: float = Field(
        0.5,
        ge=0,
        le=1,
        description='The effective area for air flow at the top of the shade, divided by'
        ' the horizontal area between glass and shade.'
    )

    bottom_opening_multiplier: float = Field(
        0.5,
        ge=0,
        le=1,
        description='The effective area for air flow at the bottom of the shade, divided'
        ' by the horizontal area between glass and shade.'
    )

    left_opening_multiplier: float = Field(
        0.5,
        ge=0,
        le=1,
        description='The effective area for air flow at the left side of the shade,'
        ' divided by the vertical area between glass and shade.'
    )

    right_opening_multiplier: float = Field(
        0.5,
        ge=0,
        le=1,
        description='The effective area for air flow at the right side of the shade,'
        ' divided by the vertical area between glass and shade.'
    )



"""Changes associated with version Honeybee schema version 1.43.5."""


def version_1_43_5(model_dict):
    """Implement changes in a Model dict to make it compatible with version 1.43.5."""
    if 'radiance' in model_dict['properties']:
        if 'modifiers' in model_dict['properties']['radiance']:
            for mod in model_dict['properties']['radiance']['modifiers']:
                if mod['type'] != 'BSDF':
                    mod['type'] = mod['type'].capitalize()
    return model_dict



"""Changes associated with version Honeybee schema version 1.40.1."""


def version_1_40_1(model_dict):
    """Implement changes in a Model dict to make it compatible with version 1.40.1."""
    if 'energy' in model_dict['properties']:
        if 'hvacs' in model_dict['properties']['energy']:
            for hvac in model_dict['properties']['energy']['hvacs']:
                if hvac['type'] != 'IdealAirSystemAbridged':  # it's detailed
                    hvac['vintage'] = VINTAGE_MAPPER[hvac['vintage']]
                    hvac['equipment_type'] = EQUIPMENT_MAPPER[hvac['equipment_type']]
    return model_dict


# dictionary to map between old and new vintages
VINTAGE_MAPPER = {
    '90.1-2013': 'ASHRAE_2013',
    '90.1-2010': 'ASHRAE_2010',
    '90.1-2007': 'ASHRAE_2007',
    '90.1-2004': 'ASHRAE_2004',
    'DOE Ref 1980-2004': 'DOE_Ref_1980_2004',
    'DOE Ref Pre-1980': 'DOE_Ref_Pre_1980'
}

# dictionary to map between the old and new HVAC equipment enumerations
EQUIPMENT_MAPPER = {
    'VAV chiller with gas boiler reheat': 'VAV_Chiller_Boiler',
    'VAV chiller with central air source heat pump reheat': 'VAV_Chiller_ASHP',
    'VAV chiller with district hot water reheat': 'VAV_Chiller_DHW',
    'VAV chiller with PFP boxes': 'VAV_Chiller_PFP',
    'VAV chiller with gas coil reheat': 'VAV_Chiller_GasCoil',
    'VAV air-cooled chiller with gas boiler reheat': 'VAV_ACChiller_Boiler',
    'VAV air-cooled chiller with central air source heat pump reheat': 'VAV_ACChiller_ASHP',
    'VAV air-cooled chiller with district hot water reheat': 'VAV_ACChiller_DHW',
    'VAV air-cooled chiller with PFP boxes': 'VAV_ACChiller_PFP',
    'VAV air-cooled chiller with gas coil reheat': 'VAV_ACChiller_GasCoil',
    'VAV district chilled water with gas boiler reheat': 'VAV_DCW_Boiler',
    'VAV district chilled water with central air source heat pump reheat': 'VAV_DCW_ASHP',
    'VAV district chilled water with district hot water reheat': 'VAV_DCW_DHW',
    'VAV district chilled water with PFP boxes': 'VAV_DCW_PFP',
    'VAV district chilled water with gas coil reheat': 'VAV_DCW_GasCoil',
    'PVAV with gas boiler reheat': 'PVAV_Boiler',
    'PVAV with central air source heat pump reheat': 'PVAV_ASHP',
    'PVAV with district hot water reheat': 'PVAV_DHW',
    'PVAV with PFP boxes': 'PVAV_PFP',
    'PVAV with gas heat with electric reheat': 'PVAV_BoilerElectricReheat',
    'PSZ-AC with baseboard electric': 'PSZAC_ElectricBaseboard',
    'PSZ-AC with baseboard gas boiler': 'PSZAC_BoilerBaseboard',
    'PSZ-AC with baseboard district hot water': 'PSZAC_DHWBaseboard',
    'PSZ-AC with gas unit heaters': 'PSZAC_GasHeaters',
    'PSZ-AC with electric coil': 'PSZAC_ElectricCoil',
    'PSZ-AC with gas coil': 'PSZAC_GasCoil',
    'PSZ-AC with gas boiler': 'PSZAC_Boiler',
    'PSZ-AC with central air source heat pump': 'PSZAC_ASHP',
    'PSZ-AC with district hot water': 'PSZAC_DHW',
    'PSZ-AC with no heat': 'PSZAC',
    'PSZ-AC district chilled water with baseboard electric': 'PSZAC_DCW_ElectricBaseboard',
    'PSZ-AC district chilled water with baseboard gas boiler': 'PSZAC_DCW_BoilerBaseboard',
    'PSZ-AC district chilled water with gas unit heaters': 'PSZAC_DCW_GasHeaters',
    'PSZ-AC district chilled water with electric coil': 'PSZAC_DCW_ElectricCoil',
    'PSZ-AC district chilled water with gas coil': 'PSZAC_DCW_GasCoil',
    'PSZ-AC district chilled water with gas boiler': 'PSZAC_DCW_Boiler',
    'PSZ-AC district chilled water with central air source heat pump': 'PSZAC_DCW_ASHP',
    'PSZ-AC district chilled water with district hot water': 'PSZAC_DCW_DHW',
    'PSZ-AC district chilled water with no heat': 'PSZAC_DCW',
    'PSZ-HP': 'PSZHP',
    'PTAC with baseboard electric': 'PTAC_ElectricBaseboard',
    'PTAC with baseboard gas boiler': 'PTAC_BoilerBaseboard',
    'PTAC with baseboard district hot water': 'PTAC_DHWBaseboard',
    'PTAC with gas unit heaters': 'PTAC_GasHeaters',
    'PTAC with electric coil': 'PTAC_ElectricCoil',
    'PTAC with gas coil': 'PTAC_GasCoil',
    'PTAC with gas boiler': 'PTAC_Boiler',
    'PTAC with central air source heat pump': 'PTAC_ASHP',
    'PTAC with district hot water': 'PTAC_DHW',
    'PTAC with no heat': 'PTAC',
    'PTHP': 'PTHP',
    'Forced air furnace': 'Furnace',
    'DOAS with fan coil chiller with boiler': 'DOAS_FCU_Chiller_Boiler',
    'DOAS with fan coil chiller with central air source heat pump': 'DOAS_FCU_Chiller_ASHP',
    'DOAS with fan coil chiller with district hot water': 'DOAS_FCU_Chiller_DHW',
    'DOAS with fan coil chiller with baseboard electric': 'DOAS_FCU_Chiller_ElectricBaseboard',
    'DOAS with fan coil chiller with gas unit heaters': 'DOAS_FCU_Chiller_GasHeaters',
    'DOAS with fan coil chiller with no heat': 'DOAS_FCU_Chiller',
    'DOAS with fan coil air-cooled chiller with boiler': 'DOAS_FCU_ACChiller_Boiler',
    'DOAS with fan coil air-cooled chiller with central air source heat pump': 'DOAS_FCU_ACChiller_ASHP',
    'DOAS with fan coil air-cooled chiller with district hot water': 'DOAS_FCU_ACChiller_DHW',
    'DOAS with fan coil air-cooled chiller with baseboard electric': 'DOAS_FCU_ACChiller_ElectricBaseboard',
    'DOAS with fan coil air-cooled chiller with gas unit heaters': 'DOAS_FCU_ACChiller_GasHeaters',
    'DOAS with fan coil air-cooled chiller with no heat': 'DOAS_FCU_ACChiller',
    'DOAS with fan coil district chilled water with boiler': 'DOAS_FCU_DCW_Boiler',
    'DOAS with fan coil district chilled water with central air source heat pump': 'DOAS_FCU_DCW_ASHP',
    'DOAS with fan coil district chilled water with district hot water': 'DOAS_FCU_DCW_DHW',
    'DOAS with fan coil district chilled water with baseboard electric': 'DOAS_FCU_DCW_ElectricBaseboard',
    'DOAS with fan coil district chilled water with gas unit heaters': 'DOAS_FCU_DCW_GasHeaters',
    'DOAS with fan coil district chilled water with no heat': 'DOAS_FCU_DCW',
    'DOAS with VRF': 'DOAS_VRF',
    'DOAS with water source heat pumps fluid cooler with boiler': 'DOAS_WSHP_FluidCooler_Boiler',
    'DOAS with water source heat pumps cooling tower with boiler': 'DOAS_WSHP_CoolingTower_Boiler',
    'DOAS with water source heat pumps with ground source heat pump': 'DOAS_WSHP_GSHP',
    'DOAS with water source heat pumps district chilled water with district hot water': 'DOAS_WSHP_DCW_DHW',
    'Baseboard electric': 'ElectricBaseboard',
    'Baseboard gas boiler': 'BoilerBaseboard',
    'Baseboard central air source heat pump': 'ASHPBaseboard',
    'Baseboard district hot water': 'DHWBaseboard',
    'Direct evap coolers with baseboard electric': 'EvapCoolers_ElectricBaseboard',
    'Direct evap coolers with baseboard gas boiler': 'EvapCoolers_BoilerBaseboard',
    'Direct evap coolers with baseboard central air source heat pump': 'EvapCoolers_ASHPBaseboard',
    'Direct evap coolers with baseboard district hot water': 'EvapCoolers_DHWBaseboard',
    'Direct evap coolers with forced air furnace': 'EvapCoolers_Furnace',
    'Direct evap coolers with gas unit heaters': 'EvapCoolers_UnitHeaters',
    'Direct evap coolers with no heat': 'EvapCoolers',
    'Fan coil chiller with boiler': 'FCU_Chiller_Boiler',
    'Fan coil chiller with central air source heat pump': 'FCU_Chiller_ASHP',
    'Fan coil chiller with district hot water': 'FCU_Chiller_DHW',
    'Fan coil chiller with baseboard electric': 'FCU_Chiller_ElectricBaseboard',
    'Fan coil chiller with gas unit heaters': 'FCU_Chiller_GasHeaters',
    'Fan coil chiller with no heat': 'FCU_Chiller',
    'Fan coil air-cooled chiller with boiler': 'FCU_ACChiller_Boiler',
    'Fan coil air-cooled chiller with central air source heat pump': 'FCU_ACChiller_ASHP',
    'Fan coil air-cooled chiller with district hot water': 'FCU_ACChiller_DHW',
    'Fan coil air-cooled chiller with baseboard electric': 'FCU_ACChiller_ElectricBaseboard',
    'Fan coil air-cooled chiller with gas unit heaters': 'FCU_ACChiller_GasHeaters',
    'Fan coil air-cooled chiller with no heat': 'FCU_ACChiller',
    'Fan coil district chilled water with boiler': 'FCU_DCW_Boiler',
    'Fan coil district chilled water with central air source heat pump': 'FCU_DCW_ASHP',
    'Fan coil district chilled water with district hot water': 'FCU_DCW_DHW',
    'Fan coil district chilled water with baseboard electric': 'FCU_DCW_ElectricBaseboard',
    'Fan coil district chilled water with gas unit heaters': 'FCU_DCW_GasHeaters',
    'Fan coil district chilled water with no heat': 'FCU_DCW',
    'Gas unit heaters': 'GasHeaters',
    'Residential AC with baseboard electric': 'ResidentialAC_ElectricBaseboard',
    'Residential AC with baseboard gas boiler': 'ResidentialAC_BoilerBaseboard',
    'Residential AC with baseboard central air source heat pump': 'ResidentialAC_ASHPBaseboard',
    'Residential AC with baseboard district hot water': 'ResidentialAC_DHWBaseboard',
    'Residential AC with residential forced air furnace': 'ResidentialAC_ResidentialFurnace',
    'Residential AC with no heat': 'ResidentialAC',
    'Residential heat pump': 'ResidentialHP',
    'Residential heat pump with no cooling': 'ResidentialHPNoCool',
    'Residential forced air furnace': 'ResidentialFurnace',
    'VRF': 'VRF',
    'Water source heat pumps fluid cooler with boiler': 'WSHP_FluidCooler_Boiler',
    'Water source heat pumps cooling tower with boiler': 'WSHP_CoolingTower_Boiler',
    'Water source heat pumps with ground source heat pump': 'WSHP_GSHP',
    'Water source heat pumps district chilled water with district hot water': 'WSHP_DCW_DHW',
    'Window AC with baseboard electric': 'WindowAC_ElectricBaseboard',
    'Window AC with baseboard gas boiler': 'WindowAC_BoilerBaseboard',
    'Window AC with baseboard central air source heat pump': 'WindowAC_ASHPBaseboard',
    'Window AC with baseboard district hot water': 'WindowAC_DHWBaseboard',
    'Window AC with forced air furnace': 'WindowAC_Furnace',
    'Window AC with unit heaters': 'WindowAC_GasHeaters',
    'Window AC with no heat': 'WindowAC'
}



"""Changes associated with version Honeybee schema version 1.43.1."""

NEW_ABRIDGED_CLASSES = ('FCUwithDOAS', 'VRFwithDOAS', 'WSHPwithDOAS')


def version_1_43_1(model_dict):
    """Implement changes in a Model dict to make it compatible with version 1.43.1."""
    if 'energy' in model_dict['properties']:
        if 'hvacs' in model_dict['properties']['energy']:
            for hvac in model_dict['properties']['energy']['hvacs']:
                if hvac['type'] in NEW_ABRIDGED_CLASSES:
                    hvac['type'] = '{}Abridged'.format(hvac['type'])
                if 'economizer_type' in hvac and hvac['economizer_type'] == 'Inferred':
                    hvac['economizer_type'] = 'NoEconomizer'
                for prop in ('sensible_heat_recovery', 'latent_heat_recovery'):
                    if prop in hvac and not isinstance(hvac[prop], (float, int)):
                        hvac[prop] = 0
    return model_dict



"""HVAC systems with DOAS, separating ventilation and meeting thermal demand."""
from pydantic import Field, constr
from enum import Enum

from ._template import _TemplateSystem, RadiantFaceTypes


class _DOASBase(_TemplateSystem):
    """Base class for Dedicated Outdoor Air System (DOAS) HVACs.

    DOAS systems separate minimum ventilation supply from the satisfaction of heating
    + cooling demand. Ventilation air tends to be supplied at neutral temperatures
    (close to room air temperature) and heating / cooling loads are met with additional
    pieces of zone equipment (eg. Fan Coil Units (FCUs)).

    Because DOAS systems only have to cool down and re-heat the minimum ventilation air,
    they tend to use less energy than all-air systems. They also tend to use less energy
    to distribute heating + cooling by pumping around hot/cold water or refrigerant
    instead of blowing hot/cold air. However, they do not provide as good of control
    over humidity and so they may not be appropriate for rooms with high latent loads
    like auditoriums, kitchens, laundromats, etc.
    """

    sensible_heat_recovery: float = Field(
        0,
        ge=0,
        le=1,
        description='A number between 0 and 1 for the effectiveness of sensible '
        'heat recovery within the system.'
    )

    latent_heat_recovery: float = Field(
        0,
        ge=0,
        le=1,
        description='A number between 0 and 1 for the effectiveness of latent '
        'heat recovery within the system.'
    )

    demand_controlled_ventilation: bool = Field(
        False,
        description='Boolean to note whether demand controlled ventilation should be '
        'used on the system, which will vary the amount of ventilation air according '
        'to the occupancy schedule of the Rooms.'
    )

    doas_availability_schedule: str = Field(
        None,
        min_length=1,
        max_length=100,
        description='An optional On/Off discrete schedule to set when the dedicated '
        'outdoor air system (DOAS) shuts off. This will not only prevent any outdoor '
        'air from flowing thorough the system but will also shut off the fans, which '
        'can result in more energy savings when spaces served by the DOAS are '
        'completely unoccupied. If None, the DOAS will be always on.'
    )


class FCUwithDOASEquipmentType(str, Enum):
    fcu_chill_gb = 'DOAS_FCU_Chiller_Boiler'
    fcu_chill_ashp = 'DOAS_FCU_Chiller_ASHP'
    fcu_chill_dhw = 'DOAS_FCU_Chiller_DHW'
    fcu_chill_base = 'DOAS_FCU_Chiller_ElectricBaseboard'
    fcu_chill_guh = 'DOAS_FCU_Chiller_GasHeaters'
    fcu_chill = 'DOAS_FCU_Chiller'
    fcu_ac_chill_gb = 'DOAS_FCU_ACChiller_Boiler'
    fcu_ac_chill_ashp = 'DOAS_FCU_ACChiller_ASHP'
    fcu_ac_chill_dhw = 'DOAS_FCU_ACChiller_DHW'
    fcu_ac_chill_base = 'DOAS_FCU_ACChiller_ElectricBaseboard'
    fcu_ac_chill_guh = 'DOAS_FCU_ACChiller_GasHeaters'
    fcu_ac_chill = 'DOAS_FCU_ACChiller'
    fcu_dcw_gb = 'DOAS_FCU_DCW_Boiler'
    fcu_dcw_ashp = 'DOAS_FCU_DCW_ASHP'
    fcu_dcw_dhw = 'DOAS_FCU_DCW_DHW'
    fcu_dcw_base = 'DOAS_FCU_DCW_ElectricBaseboard'
    fcu_dcw_guh = 'DOAS_FCU_DCW_GasHeaters'
    fcu_dcw = 'DOAS_FCU_DCW'


class WSHPwithDOASEquipmentType(str, Enum):
    wshp_fc_gb = 'DOAS_WSHP_FluidCooler_Boiler'
    wshp_ct_gb = 'DOAS_WSHP_CoolingTower_Boiler'
    wshp_gshp = 'DOAS_WSHP_GSHP'
    wshp_dcw_dhw = 'DOAS_WSHP_DCW_DHW'


class VRFwithDOASEquipmentType(str, Enum):
    vrf = 'DOAS_VRF'


class RadiantwithDOASEquipmentType(str, Enum):
    radiant_chill_gb = 'DOAS_Radiant_Chiller_Boiler'
    radiant_chill_ashp = 'DOAS_Radiant_Chiller_ASHP'
    radiant_chill_dhw = 'DOAS_Radiant_Chiller_DHW'
    radiant_ac_chill_gb = 'DOAS_Radiant_ACChiller_Boiler'
    radiant_ac_chill_ashp = 'DOAS_Radiant_ACChiller_ASHP'
    radiant_ac_chill_dhw = 'DOAS_Radiant_ACChiller_DHW'
    radiant_dcw_gb = 'DOAS_Radiant_DCW_Boiler'
    radiant_dcw_ashp = 'DOAS_Radiant_DCW_ASHP'
    radiant_dcw_dhw = 'DOAS_Radiant_DCW_DHW'


class FCUwithDOASAbridged(_DOASBase):
    """Fan Coil Unit (FCU) with DOAS HVAC system.

    All rooms/zones in the system are connected to a Dedicated Outdoor Air System
    (DOAS) that supplies a constant volume of ventilation air at the same temperature
    to all rooms/zones. The ventilation air temperature will vary from 21.1C (70F)
    to 15.5C (60F) depending on the outdoor air temperature (the DOAS supplies cooler air
    when outdoor conditions are warmer). The ventilation air temperature is maintained
    by a chilled water cooling coil and a heating coil. The heating coil is a hot
    water coil except when electric baseboards or gas heaters are specified, in
    which case the heating coil is a single-speed direct expansion (DX) heat pump
    with a backup electrical resistance coil.

    Each room/zone also receives its own Fan Coil Unit (FCU), which meets the heating
    and cooling loads of the space. The cooling coil in the FCU is always chilled
    water cooling coil, which is connected to a chilled water loop operating
    at 6.7C (44F). The heating coil is a hot water coil except when when electric
    baseboards or gas heaters are specified. Hot water temperature is 82C (180F) for
    boiler/district heating and 49C (120F) when ASHP is used.

    The FCU with DOAS template is relatively close in performance to active chilled
    beams (ACBs). When using this template to represent ACBs, care must be taken
    to ensure that the DOAS ventilation air requirement is sufficient to extract
    the heating cooling from the ACB. If so, then this FCUwithDOAS template can be
    used but with the energy use of the FCU fans ignored.
    """

    type: constr(regex='^FCUwithDOASAbridged$') = 'FCUwithDOASAbridged'

    equipment_type: FCUwithDOASEquipmentType = Field(
        FCUwithDOASEquipmentType.fcu_chill_gb,
        description='Text for the specific type of system equipment from the '
        'FCUwithDOASEquipmentType enumeration.'
    )


class WSHPwithDOASAbridged(_DOASBase):
    """Water Source Heat Pump (WSHP) with DOAS HVAC system.

    All rooms/zones in the system are connected to a Dedicated Outdoor Air System
    (DOAS) that supplies a constant volume of ventilation air at the same temperature
    to all rooms/zones. The ventilation air temperature will vary from 21.1C (70F)
    to 15.5C (60F) depending on the outdoor air temperature (the DOAS supplies cooler air
    when outdoor conditions are warmer). The ventilation air temperature is maintained
    by a chilled water cooling coil and a hot water heating coil except when the
    ground source heat pump (GSHP) option is selected. In this case, the ventilation
    air temperature is maintained by a two-speed direct expansion (DX) cooling coil
    and a single-speed DX heating coil with backup electrical resistance heat.

    Each room/zone also receives its own Water Source Heat Pump (WSHP), which meets
    the heating and cooling loads of the space. All WSHPs are connected to the
    same water condenser loop, which has its temperature maintained by the
    equipment_type (eg. Boiler with Cooling Tower).
    """

    type: constr(regex='^WSHPwithDOASAbridged$') = 'WSHPwithDOASAbridged'

    equipment_type: WSHPwithDOASEquipmentType = Field(
        WSHPwithDOASEquipmentType.wshp_fc_gb,
        description='Text for the specific type of system equipment from the '
        'WSHPwithDOASEquipmentType enumeration.'
    )


class VRFwithDOASAbridged(_DOASBase):
    """Variable Refrigerant Flow (VRF) with DOAS HVAC system.

    All rooms/zones in the system are connected to a Dedicated Outdoor Air System
    (DOAS) that supplies a constant volume of ventilation air at the same temperature
    to all rooms/zones. The ventilation air temperature will vary from 21.1C (70F)
    to 15.5C (60F) depending on the outdoor air temperature (the DOAS supplies cooler air
    when outdoor conditions are warmer). The ventilation air temperature is maintained
    by a single speed direct expansion (DX) cooling coil along with a single-speed
    direct expansion (DX) heat pump with a backup electrical resistance coil.

    Each room/zone also receives its own Variable Refrigerant Flow (VRF) terminal,
    which meets the heating and cooling loads of the space. All room/zone terminals
    are connected to the same outdoor unit, meaning that either all rooms must be
    in cooling or heating mode together.
    """

    type: constr(regex='^VRFwithDOASAbridged$') = 'VRFwithDOASAbridged'

    equipment_type: VRFwithDOASEquipmentType = Field(
        VRFwithDOASEquipmentType.vrf,
        description='Text for the specific type of system equipment from the '
        'VRFwithDOASEquipmentType enumeration.'
    )


class RadiantwithDOASAbridged(_DOASBase):
    """Low Temperature Radiant with DOAS HVAC system.

    This HVAC template will change the floor and/or ceiling constructions
    of the Rooms that it is applied to, replacing them with a construction that
    aligns with the radiant_type property (eg. CeilingMetalPanel).

    All rooms/zones in the system are connected to a Dedicated Outdoor Air System
    (DOAS) that supplies a constant volume of ventilation air at the same temperature
    to all rooms/zones. The ventilation air temperature will vary from 21.1C (70F)
    to 15.5C (60F) depending on the outdoor air temperature (the DOAS supplies cooler air
    when outdoor conditions are warmer). The ventilation air temperature is maintained
    by a two-speed direct expansion (DX) cooling coil and a single-speed DX
    heating coil with backup electrical resistance heat.

    The heating and cooling needs of the space are met with the radiant constructions,
    which use chilled water at 12.8C (55F) and a hot water temperature somewhere
    between 32.2C (90F) and 49C (120F) (warmer temperatures are used in colder
    climate zones).

    Note that radiant systems are particularly limited in cooling capacity and
    using them may result in many unmet hours. To reduce unmet hours, one can
    remove carpets, reduce internal loads, reduce solar and envelope gains during
    peak times, add thermal mass, and use an expanded comfort range.
    """

    type: constr(regex='^RadiantwithDOASAbridged$') = 'RadiantwithDOASAbridged'

    equipment_type: RadiantwithDOASEquipmentType = Field(
        RadiantwithDOASEquipmentType.radiant_chill_gb,
        description='Text for the specific type of system equipment from the '
        'RadiantwithDOASEquipmentType enumeration.'
    )

    radiant_face_type: RadiantFaceTypes = Field(
        RadiantFaceTypes.floor,
        description='Text to indicate which faces are thermally active by default. '
        'Note that this property has no effect when the rooms to which the HVAC '
        'system is assigned have constructions with internal source materials. '
        'In this case, those constructions will dictate the thermally active '
        'surfaces.'
    )

    minimum_operation_time: float = Field(
        1.0,
        gt=0,
        description='A number for the minimum number of hours of operation '
        'for the radiant system before it shuts off.'
    )

    switch_over_time: float = Field(
        24.0,
        gt=0,
        description='A number for the minimum number of hours for when the system '
        'can switch between heating and cooling.'
    )



"""Ideal Air Schema"""
from pydantic import Field, constr, confloat
from typing import Union
from enum import Enum

from ._base import IDdEnergyBaseModel
from ..altnumber import Autocalculate


class SHWEquipmentType(str, Enum):
    gas_waterheater = 'Gas_WaterHeater'
    electric_waterheater = 'Electric_WaterHeater'
    heatpump_waterheater = 'HeatPump_WaterHeater'
    gas_tanklessheater = 'Gas_TanklessHeater'
    electric_tanklessheater = 'Electric_TanklessHeater'


class SHWSystem(IDdEnergyBaseModel):
    """Provides a model for a Service Hot Water system."""

    type: constr(regex='^SHWSystem$') = 'SHWSystem'

    equipment_type: SHWEquipmentType = Field(
        SHWEquipmentType.gas_waterheater,
        description='Text to indicate the type of air-side economizer used on the '
        'ideal air system. Economizers will mix in a greater amount of outdoor '
        'air to cool the zone (rather than running the cooling system) when '
        'the zone needs cooling and the outdoor air is cooler than the zone.'
    )

    heater_efficiency: Union[confloat(gt=0), Autocalculate] = Field(
        Autocalculate(),
        description='A number for the efficiency of the heater within the system. '
        'For Gas systems, this is the efficiency of the burner. For HeatPump '
        'systems, this is the rated COP of the system. For electric systems, this '
        'should usually be set to 1. If set to Autocalculate, this value will '
        'automatically be set based on the equipment_type. Gas_WaterHeater - 0.8, '
        'Electric_WaterHeater - 1.0, HeatPump_WaterHeater - 3.5, Gas_TanklessHeater '
        '- 0.8, Electric_TanklessHeater - 1.0.'
    )

    ambient_condition: Union[float, str] = Field(
        22,
        description='A number for the ambient temperature in which the hot '
        'water tank is located [C]. This can also be the identifier of a Room '
        'in which the tank is located.'
    )

    ambient_loss_coefficient: float = Field(
        6,
        description='A number for the loss of heat from the water heater '
        'tank to the surrounding ambient conditions [W/K].'
    )



"""Simulation Parameter Schema"""
from pydantic import Field, validator, root_validator, constr, conlist
from typing import List
from enum import Enum

from .._base import NoExtraBaseModel
from ._base import DatedBaseModel
from .designday import DesignDay


class ReportingFrequency(str, Enum):
    timestep = 'Timestep'
    hourly = 'Hourly'
    daily = 'Daily'
    monthly = 'Monthly'
    annual = 'Annual'


class SimulationOutput(NoExtraBaseModel):
    """Lists the outputs to report from the simulation and their format."""

    type: constr(regex='^SimulationOutput$') = 'SimulationOutput'

    reporting_frequency: ReportingFrequency = ReportingFrequency.hourly

    outputs: List[str] = Field(
        default=None,
        description='A list of EnergyPlus output names as strings, which are requested '
        'from the simulation.'
    )

    summary_reports: List[str] = Field(
        default=None,
        description='A list of EnergyPlus summary report names as strings.'
    )

    unmet_setpoint_tolerance: float = Field(
        1.11,
        ge=0,
        le=10,
        description='A number in degrees Celsius for the difference that '
        'the zone conditions must be from the thermostat setpoint in order '
        'for the setpoint to be considered unmet. This will affect how unmet '
        'hours are reported in the output. ASHRAE 90.1 uses a tolerance of '
        '1.11C, which is equivalent to 1.8F.'
    )


class SimulationControl(NoExtraBaseModel):
    """Used to specify which types of calculations to run."""

    type: constr(regex='^SimulationControl$') = 'SimulationControl'

    do_zone_sizing: bool = Field(
        default=True,
        description='Boolean for whether the zone sizing calculation should be run.'
    )

    do_system_sizing: bool = Field(
        default=True,
        description='Boolean for whether the system sizing calculation should be run.'
    )

    do_plant_sizing: bool = Field(
        default=True,
        description='Boolean for whether the plant sizing calculation should be run.'
    )

    run_for_run_periods: bool = Field(
        default=True,
        description='Boolean for whether the simulation should be run for the '
        'sizing periods.'
    )

    run_for_sizing_periods: bool = Field(
        default=False,
        description='Boolean for whether the simulation should be run for the '
        'run periods.'
    )


class SolarDistribution(str, Enum):
    minimal_shadowing = 'MinimalShadowing'
    full_exterior = 'FullExterior'
    full_interior_and_exterior = 'FullInteriorAndExterior'
    full_exterior_with_reflection = 'FullExteriorWithReflections'
    full_interior_and_exterior_with_reflections = \
        'FullInteriorAndExteriorWithReflections'


class CalculationMethod(str, Enum):
    polygon_clipping = 'PolygonClipping'
    pixel_counting = 'PixelCounting'


class CalculationUpdateMethod(str, Enum):
    periodic = 'Periodic'
    timestep = 'Timestep'


class ShadowCalculation(NoExtraBaseModel):
    """Used to describe settings for EnergyPlus shadow calculation."""

    type: constr(regex='^ShadowCalculation$') = 'ShadowCalculation'

    solar_distribution: SolarDistribution = \
        SolarDistribution.full_exterior_with_reflection

    calculation_method: CalculationMethod = Field(
        CalculationMethod.polygon_clipping,
        description='Text noting whether CPU-based polygon clipping method or'
        'GPU-based pixel counting method should be used. For low numbers of shading'
        'surfaces (less than ~200), PolygonClipping requires less runtime than'
        'PixelCounting. However, PixelCounting runtime scales significantly'
        'better at higher numbers of shading surfaces. PixelCounting also has'
        'no limitations related to zone concavity when used with any'
        '“FullInterior” solar distribution options.'
    )

    calculation_update_method: CalculationUpdateMethod = Field(
        CalculationUpdateMethod.periodic,
        description='Text describing how often the solar and shading calculations '
        'are updated with respect to the flow of time in the simulation.'
    )

    calculation_frequency: int = Field(
        30,
        ge=1,
        description='Integer for the number of days in each period for which a '
        'unique shadow calculation will be performed. This field is only used '
        'if the Periodic calculation_method is used.'
    )

    maximum_figures: int = Field(
        15000,
        ge=200,
        description='Number of allowable figures in shadow overlap calculations.'
    )


class DaysOfWeek(str, Enum):
    sunday = 'Sunday'
    monday = 'Monday'
    tuesday = 'Tuesday'
    wednesday = 'Wednesday'
    thursday = 'Thursday'
    friday = 'Friday'
    saturday = 'Saturday'


class DaylightSavingTime(DatedBaseModel):
    """Used to describe the daylight savings time for the simulation."""

    type: constr(regex='^DaylightSavingTime$') = 'DaylightSavingTime'

    start_date: List[int] = Field(
        [3, 12],
        min_items=2,
        max_items=3,
        description='A list of two integers for [month, day], representing the date '
        'for the start of daylight savings time. Default: 12 Mar (daylight savings '
        'in the US in 2017).'
    )

    @validator('start_date')
    def check_start_date(cls, v):
        return cls.check_date(v)

    end_date: List[int] = Field(
        [11, 5],
        min_items=2,
        max_items=3,
        description='A list of two integers for [month, day], representing the date '
        'for the end of daylight savings time. Default: 5 Nov (daylight savings '
        'in the US in 2017).'
    )

    @validator('end_date')
    def check_end_date(cls, v):
        return cls.check_date(v)


class RunPeriod(DatedBaseModel):
    """Used to describe the time period over which to run the simulation."""

    type: constr(regex='^RunPeriod$') = 'RunPeriod'

    start_date: List[int] = Field(
        [1, 1],
        min_items=2,
        max_items=2,
        description='A list of two integers for [month, day], representing the date '
        'for the start of the run period. Must be before the end date.'
    )

    end_date: List[int] = Field(
        [12, 31],
        min_items=2,
        max_items=2,
        description='A list of two integers for [month, day], representing the date '
        'for the end of the run period. Must be after the start date.'
    )

    start_day_of_week: DaysOfWeek = Field(
        default=DaysOfWeek.sunday,
        description='Text for the day of the week on which the simulation starts.'
    )

    holidays: List[conlist(int, min_items=2, max_items=2)] = Field(
        default=None,
        description='A list of lists where each sub-list consists of two integers '
        'for [month, day], representing a date which is a holiday within the '
        'simulation. If None, no holidays are applied.'
    )

    daylight_saving_time: DaylightSavingTime = Field(
        default=None,
        description='A DaylightSavingTime to dictate the start and end dates '
        'of daylight saving time. If None, no daylight saving time is applied '
        'to the simulation.'
    )

    leap_year: bool = Field(
        default=False,
        description='Boolean noting whether the simulation will be run for a leap year.'
    )

    @root_validator
    def check_dates(cls, values):
        """Check that all of the input dates are valid."""
        start_date = values.get('start_date')
        end_date = values.get('end_date')
        holidays = values.get('holidays')
        leap_year = values.get('leap_year')
        cls.check_date(start_date, leap_year)
        cls.check_date(end_date, leap_year)
        if holidays is not None:
            for hol in holidays:
                cls.check_date(hol, leap_year)
        return values


class EfficiencyStandards(str, Enum):
    ashrae_2019 = 'ASHRAE_2019'
    ashrae_2016 = 'ASHRAE_2016'
    ashrae_2013 = 'ASHRAE_2013'
    ashrae_2010 = 'ASHRAE_2010'
    ashrae_2007 = 'ASHRAE_2007'
    ashrae_2004 = 'ASHRAE_2004'
    doe_ref_1980_2004 = 'DOE_Ref_1980_2004'
    doe_ref_pre_1980 = 'DOE_Ref_Pre_1980'


class ClimateZones(str, Enum):
    zone_0A = '0A'
    zone_1A = '1A'
    zone_2A = '2A'
    zone_3A = '3A'
    zone_4A = '4A'
    zone_5A = '5A'
    zone_6A = '6A'
    zone_0B = '0B'
    zone_1B = '1B'
    zone_2B = '2B'
    zone_3B = '3B'
    zone_4B = '4B'
    zone_5B = '5B'
    zone_6B = '6B'
    zone_3C = '3C'
    zone_4C = '4C'
    zone_5C = '5C'
    zone_7 = '7'
    zone_8 = '8'


class BuildingTypes(str, Enum):
    Residential = 'Residential'
    NonResidential = 'NonResidential'
    MidriseApartment = 'MidriseApartment'
    HighriseApartment = 'HighriseApartment'
    LargeOffice = 'LargeOffice'
    MediumOffice = 'MediumOffice'
    SmallOffice = 'SmallOffice'
    Retail = 'Retail'
    StripMall = 'StripMall'
    PrimarySchool = 'PrimarySchool'
    SecondarySchool = 'SecondarySchool'
    SmallHotel = 'SmallHotel'
    LargeHotel = 'LargeHotel'
    Hospital = 'Hospital'
    Outpatient = 'Outpatient'
    Warehouse = 'Warehouse'
    SuperMarket = 'SuperMarket'
    FullServiceRestaurant = 'FullServiceRestaurant'
    QuickServiceRestaurant = 'QuickServiceRestaurant'
    Laboratory = 'Laboratory'
    Courthouse = 'Courthouse'


class SizingParameter(NoExtraBaseModel):
    """Used to specify heating and cooling sizing criteria and safety factors."""

    type: constr(regex='^SizingParameter$') = 'SizingParameter'

    design_days: List[DesignDay] = Field(
        default=None,
        description='A list of DesignDays that represent the criteria for which '
        'the HVAC systems will be sized.'
    )

    heating_factor: float = Field(
        1.25,
        gt=0,
        description='A number that will be multiplied by the peak heating load'
        ' for each zone in order to size the heating system.'
    )

    cooling_factor: float = Field(
        1.15,
        gt=0,
        description='A number that will be multiplied by the peak cooling load'
        ' for each zone in order to size the heating system.'
    )

    efficiency_standard: EfficiencyStandards = Field(
        default=None,
        description='Text to specify the efficiency standard, which will '
        'automatically set the efficiencies of all HVAC equipment when provided. '
        'Note that providing a standard here will cause the OpenStudio translation '
        'process to perform an additional sizing calculation with EnergyPlus, '
        'which is needed since the default efficiencies of equipment vary depending '
        'on their size. THIS WILL SIGNIFICANTLY INCREASE TRANSLATION TIME TO '
        'OPENSTUDIO. However, it is often worthwhile when the goal is to match the '
        'HVAC specification with a particular standard.'
    )

    climate_zone: ClimateZones = Field(
        default=None,
        description='Text indicating the ASHRAE climate zone to be used with the '
        'efficiency_standard. When unspecified, the climate zone will be inferred from '
        'the design days on this sizing parameter object.'
    )

    building_type: str = Field(
        default=None,
        description='Text for the building type to be used in the efficiency_standard. '
        'If the type is not recognized or is None, it will be assumed that the building '
        'is a generic NonResidential. The following have specified systems per the '
        'standard:  Residential, NonResidential, MidriseApartment, HighriseApartment, '
        'LargeOffice, MediumOffice, SmallOffice, Retail, StripMall, '
        'PrimarySchool, SecondarySchool, SmallHotel, LargeHotel, Hospital, Outpatient, '
        'Warehouse, SuperMarket, FullServiceRestaurant, QuickServiceRestaurant, '
        'Laboratory, Courthouse.'
    )

    bypass_efficiency_sizing: bool = Field(
        default=False,
        description='A boolean to indicate whether the efficiency standard should '
        'trigger an sizing run that sets the efficiencies of all HVAC equipment in '
        'the Model (False) or the standard should only be written into the OSM and '
        'the sizing run should be bypassed (True). Bypassing the sizing run is useful '
        'when you only want to check that the overall HVAC system architecture is '
        'correct and you do not want to wait the extra time that it takes to run the '
        'sizing calculation.'
    )


class TerrianTypes(str, Enum):
    ocean = 'Ocean'
    country = 'Country'
    suburbs = 'Suburbs'
    urban = 'Urban'
    city = 'City'


class SimulationParameter(NoExtraBaseModel):
    """The complete set of EnergyPlus Simulation Settings."""

    type: constr(regex='^SimulationParameter$') = 'SimulationParameter'

    output: SimulationOutput = Field(
        default=None,
        description='A SimulationOutput that lists the desired outputs from the '
        'simulation and the format in which to report them.'
    )

    run_period: RunPeriod = Field(
        default=None,
        description='A RunPeriod to describe the time period over which to '
        'run the simulation.'
    )

    timestep: int = Field(
        default=6,
        ge=1,
        le=60,
        description='An integer for the number of timesteps per hour at which the '
        'energy calculation will be run.'
    )

    @validator('timestep')
    def check_timestep(cls, v):
        valid_timesteps = (1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 30, 60)
        assert v in valid_timesteps, \
            '"{}" is not a valid timestep. Choose from {}'.format(v, valid_timesteps)

    simulation_control: SimulationControl = Field(
        default=None,
        description='A SimulationControl object that describes which types of '
        'calculations to run.'
    )

    shadow_calculation: ShadowCalculation = Field(
        default=None,
        description='A ShadowCalculation object describing settings for the '
        'EnergyPlus Shadow Calculation.'
    )

    sizing_parameter: SizingParameter = Field(
        default=None,
        description='A SizingParameter object with criteria for sizing the '
        'heating and cooling system.'
    )

    north_angle: float = Field(
        default=0,
        ge=-360,
        le=360,
        description='A number between -360 and 360 for the north direction in degrees.'
        'This is the counterclockwise difference between the North and the '
        'positive Y-axis. 90 is West and 270 is East. Note that this '
        'is different than the convention used in EnergyPlus, which uses '
        'clockwise difference instead of counterclockwise difference.'
    )

    terrain_type: TerrianTypes = Field(
        default=TerrianTypes.city,
        description='Text for the terrain in which the model sits. This is used '
        'to determine the wind profile over the height of the rooms.'
    )



"""Programtype Schema"""
from pydantic import Field, constr

from ._base import IDdEnergyBaseModel


class InternalMassAbridged(IDdEnergyBaseModel):

    type: constr(regex='^InternalMassAbridged$') = 'InternalMassAbridged'

    construction: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='Identifier for an OpaqueConstruction that represents the '
        'material that the internal thermal mass is composed of.'
    )

    area: float = Field(
        ...,
        gt=0,
        description='A number representing the surface area of the internal mass that '
        'is exposed to the Room air. This value should always be in square '
        'meters regardless of what units system the parent model is a part of.'
    )



"""Geometry objects for model."""
from pydantic import Field, constr, conlist, conint
from typing import List
from ._base import NoExtraBaseModel


class Point3D(NoExtraBaseModel):
    """A point object in 3D space."""

    type: constr(regex='^Point3D$') = 'Point3D'

    x: float = Field(
        ...,
        description='Number for X coordinate.'
    )

    y: float = Field(
        ...,
        description='Number for Y coordinate.'
    )

    z: float = Field(
        ...,
        description='Number for Z coordinate.'
    )


class LineSegment3D(NoExtraBaseModel):
    """A single line segment face in 3D space."""

    type: constr(regex='^LineSegment3D$') = 'LineSegment3D'

    p: List[float] = Field(
        ...,
        description="Line segment base point as 3 (x, y, z) values.",
        min_items=3,
        max_items=3
    )

    v: List[float] = Field(
        ...,
        description="Line segment direction vector as 3 (x, y, z) values.",
        min_items=3,
        max_items=3
    )


class Plane(NoExtraBaseModel):

    type: constr(regex='^Plane$') = 'Plane'

    n: List[float] = Field(
        ...,
        description="Plane normal as 3 (x, y, z) values.",
        min_items=3,
        max_items=3
    )

    o: List[float] = Field(
        ...,
        description="Plane origin as 3 (x, y, z) values",
        min_items=3,
        max_items=3
    )

    x: List[float] = Field(
        default=None,
        description="Plane x-axis as 3 (x, y, z) values. If None, it is autocalculated.",
        min_items=3,
        max_items=3
    )


class Face3D(NoExtraBaseModel):
    """A single planar face in 3D space."""

    type: constr(regex='^Face3D$') = 'Face3D'

    boundary: List[conlist(float, min_items=3, max_items=3)] = Field(
        ...,
        min_items=3,
        description='A list of points representing the outer boundary vertices of '
        'the face. The list should include at least 3 points and each point '
        'should be a list of 3 (x, y, z) values.'
    )

    holes: List[conlist(conlist(float, min_items=3, max_items=3), min_items=3)] = Field(
        default=None,
        description='Optional list of lists with one list for each hole in the face.'
        'Each hole should be a list of at least 3 points and each point a list '
        'of 3 (x, y, z) values. If None, it will be assumed that there are no '
        'holes in the face.'
    )

    plane: Plane = Field(
        default=None,
        description='Optional Plane indicating the plane in which the face exists.'
        'If None, the plane will usually be derived from the boundary points.'
    )


class Color(NoExtraBaseModel):
    """A RGB color."""

    type: constr(regex='^Color$') = 'Color'

    r: int = Field(
        ...,
        ge=0,
        le=255,
        description='Value for red channel.'
    )

    g: int = Field(
        ...,
        ge=0,
        le=255,
        description='Value for green channel.'
    )

    b: int = Field(
        ...,
        ge=0,
        le=255,
        description='Value for blue channel.'
    )

    a: int = Field(
        default=255,
        ge=0,
        le=255,
        description='Value for the alpha channel, which defines the opacity as a '
        'number between 0 (fully transparent) and 255 (fully opaque).'
    )


class Mesh3D(NoExtraBaseModel):
    """A mesh in 3D space."""

    type: constr(regex='^Mesh3D$') = 'Mesh3D'

    vertices: List[conlist(float, min_items=3, max_items=3)] = Field(
        ...,
        min_items=3,
        description='A list of points representing the vertices of the mesh. '
        'The list should include at least 3 points and each point '
        'should be a list of 3 (x, y, z) values.'
    )

    faces: List[conlist(conint(ge=0), min_items=3, max_items=4)] = Field(
        ...,
        min_items=1,
        description='A list of lists with each sub-list having either 3 or 4 '
        'integers. These integers correspond to indices within the list of vertices.'
    )

    colors: List[Color] = Field(
        None,
        description='An optional list of colors that correspond to either the faces '
        'of the mesh or the vertices of the mesh.'
    )



"""Base class used by radiance schema objects."""
from pydantic import Field, validator
import re
from .._base import NoExtraBaseModel


class IDdRadianceBaseModel(NoExtraBaseModel):
    """Base class for all objects requiring a valid Radiance identifier."""

    identifier: str = Field(
        ...,
        regex=r'^[.A-Za-z0-9_-]+$',
        min_length=1,
        description='Text string for a unique Radiance object. Must not contain spaces '
                    'or special characters. This will be used to identify the object '
                    'across a model and in the exported Radiance files.'
    )

    display_name: str = Field(
        default=None,
        description='Display name of the object with no character restrictions.'
    )



"""Global construction-set for Model."""
import pathlib
import json

from typing import List, Union
from pydantic import constr, Field

from honeybee_standards import energy_default

from .._base import NoExtraBaseModel
from .constructionset import WallConstructionSetAbridged, FloorConstructionSetAbridged, \
    RoofCeilingConstructionSetAbridged, ApertureConstructionSetAbridged, \
    DoorConstructionSetAbridged
from .construction import OpaqueConstructionAbridged, WindowConstructionAbridged, \
    ShadeConstruction, AirBoundaryConstructionAbridged
from .material import EnergyMaterial, EnergyMaterialNoMass, \
    EnergyWindowMaterialGlazing, EnergyWindowMaterialGas


# import constructionset default values from honeybee standards
_DEFAULTS = json.loads(pathlib.Path(energy_default).read_text())
_CSET = [
    ms for ms in _DEFAULTS['construction_sets']
    if ms['identifier'] == 'Default Generic Construction Set'][0]
_CONSTRUCTION_NAMES = [
    'Generic Exterior Wall', 'Generic Interior Wall', 'Generic Underground Wall',
    'Generic Exposed Floor', 'Generic Interior Floor', 'Generic Ground Slab',
    'Generic Roof', 'Generic Interior Ceiling', 'Generic Underground Roof',
    'Generic Double Pane', 'Generic Single Pane', 'Generic Exterior Door',
    'Generic Interior Door', 'Generic Shade', 'Generic Context', 'Generic Air Boundary'
]
_CONSTRUCTIONS = [
    OpaqueConstructionAbridged.parse_obj(m)
    if m['type'] == 'OpaqueConstructionAbridged'
    else WindowConstructionAbridged.parse_obj(m)
    if m['type'] == 'WindowConstructionAbridged'
    else ShadeConstruction.parse_obj(m)
    if m['type'] == 'ShadeConstruction'
    else AirBoundaryConstructionAbridged.parse_obj(m)
    for m in _DEFAULTS['constructions'] if m['identifier'] in _CONSTRUCTION_NAMES
]
_MATERIAL_NAMES = [
    'Generic 25mm Wood', 'Generic Clear Glass', 'Generic LW Concrete',
    'Generic Ceiling Air Gap', 'Generic Acoustic Tile', 'Generic Gypsum Board',
    'Generic Wall Air Gap', 'Generic Painted Metal', 'Generic 50mm Insulation',
    'Generic Roof Membrane', 'Generic Brick', 'Generic HW Concrete',
    'Generic Low-e Glass', 'Generic Window Air Gap', 'Generic 25mm Insulation'
]
_MATERIALS = [
    EnergyMaterial.parse_obj(m)
    if m['type'] == 'EnergyMaterial'
    else EnergyMaterialNoMass.parse_obj(m)
    if m['type'] == 'EnergyMaterialNoMass'
    else EnergyWindowMaterialGlazing.parse_obj(m)
    if m['type'] == 'EnergyWindowMaterialGlazing'
    else EnergyWindowMaterialGas.parse_obj(m)
    for m in _DEFAULTS['materials'] if m['identifier'] in _MATERIAL_NAMES
]


class GlobalConstructionSet(NoExtraBaseModel):

    type: constr(regex='^GlobalConstructionSet$') = 'GlobalConstructionSet'

    materials: List[Union[
        EnergyMaterial, EnergyMaterialNoMass,
        EnergyWindowMaterialGlazing, EnergyWindowMaterialGas
    ]] = Field(
        default=_MATERIALS,
        description='Global Honeybee Energy materials.',
        readOnly=True
    )

    constructions: List[Union[
        OpaqueConstructionAbridged, WindowConstructionAbridged,
        ShadeConstruction, AirBoundaryConstructionAbridged
    ]] = Field(
        default=_CONSTRUCTIONS,
        description='Global Honeybee Energy constructions.',
        readOnly=True
    )

    wall_set: WallConstructionSetAbridged = Field(
        default=WallConstructionSetAbridged.parse_obj(_CSET['wall_set']),
        description='Global Honeybee WallConstructionSet.',
        readOnly=True
    )

    floor_set: FloorConstructionSetAbridged = Field(
        default=FloorConstructionSetAbridged.parse_obj(_CSET['floor_set']),
        description='Global Honeybee FloorConstructionSet.',
        readOnly=True
    )

    roof_ceiling_set: RoofCeilingConstructionSetAbridged = Field(
        default=RoofCeilingConstructionSetAbridged.parse_obj(_CSET['roof_ceiling_set']),
        description='Global Honeybee RoofCeilingConstructionSet.',
        readOnly=True
    )

    aperture_set: ApertureConstructionSetAbridged = Field(
        default=ApertureConstructionSetAbridged.parse_obj(_CSET['aperture_set']),
        description='Global Honeybee ApertureConstructionSet.',
        readOnly=True
    )

    door_set: DoorConstructionSetAbridged = Field(
        default=DoorConstructionSetAbridged.parse_obj(_CSET['door_set']),
        description='Global Honeybee DoorConstructionSet.',
        readOnly=True
    )

    shade_construction: str = Field(
        default=_CSET['shade_construction'],
        description='Global Honeybee Construction for building-attached Shades.',
        readOnly=True
    )

    context_construction: str = Field(
        default='Generic Context',
        description='Global Honeybee Construction for context Shades.',
        readOnly=True
    )

    air_boundary_construction: str = Field(
        default=_CSET['air_boundary_construction'],
        description='Global Honeybee Construction for AirBoundary Faces.',
        readOnly=True
    )



"""Model energy properties."""
from pydantic import Field, constr
from typing import List, Union

from .._base import NoExtraBaseModel
from .constructionset import ConstructionSetAbridged, ConstructionSet
from .global_constructionset import GlobalConstructionSet
from .construction import OpaqueConstructionAbridged, WindowConstructionAbridged, \
    ShadeConstruction, AirBoundaryConstructionAbridged, OpaqueConstruction, \
    WindowConstruction, AirBoundaryConstruction, WindowConstructionShadeAbridged, \
    WindowConstructionShade, WindowConstructionDynamicAbridged, WindowConstructionDynamic
from .material import EnergyMaterial, EnergyMaterialNoMass, EnergyMaterialVegetation, \
    EnergyWindowMaterialSimpleGlazSys, EnergyWindowMaterialGlazing, \
    EnergyWindowMaterialGas, EnergyWindowMaterialGasCustom, \
    EnergyWindowMaterialGasMixture, EnergyWindowFrame, \
    EnergyWindowMaterialBlind, EnergyWindowMaterialShade
from .programtype import ProgramTypeAbridged, ProgramType
from .load import PeopleAbridged, LightingAbridged, ElectricEquipmentAbridged, \
    GasEquipmentAbridged, ServiceHotWaterAbridged, InfiltrationAbridged, \
    VentilationAbridged, SetpointAbridged, ProcessAbridged
from .daylight import DaylightingControl
from .ventcool import VentilationControlAbridged, VentilationFan, VentilationOpening, \
    VentilationSimulationControl, AFNCrack
from .internalmass import InternalMassAbridged
from .schedule import ScheduleTypeLimit, ScheduleRulesetAbridged, \
    ScheduleFixedIntervalAbridged, ScheduleRuleset, ScheduleFixedInterval
from .hvac.idealair import IdealAirSystemAbridged
from .hvac.allair import VAV, PVAV, PSZ, PTAC, ForcedAirFurnace
from .hvac.doas import FCUwithDOASAbridged, WSHPwithDOASAbridged, VRFwithDOASAbridged, \
    RadiantwithDOASAbridged
from .hvac.heatcool import FCU, WSHP, VRF, Baseboard, EvaporativeCooler, Residential, \
    WindowAC, GasUnitHeater, Radiant
from .hvac.detailed import DetailedHVAC
from .shw import SHWSystem
from .generator import PVProperties, ElectricLoadCenter


class ShadeMeshEnergyPropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^ShadeMeshEnergyPropertiesAbridged$') = \
        'ShadeMeshEnergyPropertiesAbridged'

    construction: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of a ShadeConstruction to set the reflectance and '
        'specularity of the Shade. If None, it will be a generic context '
        'construction that is completely diffuse with 0.2 visible and solar '
        'reflectance. Unless it is building attached, in which case it will be '
        'set by the default generic ConstructionSet.'
    )

    transmittance_schedule: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of a schedule to set the transmittance of the shade, '
        'which can vary throughout the simulation. If None, the shade will '
        'be completely opaque.'
    )


class ShadeEnergyPropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^ShadeEnergyPropertiesAbridged$') = \
        'ShadeEnergyPropertiesAbridged'

    construction: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of a ShadeConstruction to set the reflectance and '
        'specularity of the Shade. If None, the construction is set by the'
        'parent Room construction_set, the Model global_construction_set or '
        '(in the case fo an orphaned shade) the EnergyPlus default of 0.2 '
        'diffuse reflectance.'
    )

    transmittance_schedule: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of a schedule to set the transmittance of the shade, '
        'which can vary throughout the simulation. If None, the shade will '
        'be completely opaque.'
    )

    pv_properties: PVProperties = Field(
        default=None,
        description='An optional PVProperties object to specify photovoltaic '
        'behavior of the Shade. If None, the Shade will have no Photovoltaic '
        'properties. Note that the normal of the Shade is important in '
        'determining the performance of the shade as a PV geometry.'
    )


class DoorEnergyPropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^DoorEnergyPropertiesAbridged$') = \
        'DoorEnergyPropertiesAbridged'

    construction: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of an OpaqueConstruction or WindowConstruction '
        'for the door. Note that the host door must have the is_glass property '
        'set to True to assign a WindowConstruction. If None, the construction '
        'is set by the parent Room construction_set or the Model '
        'global_construction_set.'
    )

    vent_opening: VentilationOpening = Field(
        default=None,
        description='An optional VentilationOpening to specify the operable '
        'portion of the Door.'
    )


class ApertureEnergyPropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^ApertureEnergyPropertiesAbridged$') = \
        'ApertureEnergyPropertiesAbridged'

    construction: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of a WindowConstruction for the aperture. If None, the '
        'construction is set by the parent Room construction_set or the Model '
        'global_construction_set.'
    )

    vent_opening: VentilationOpening = Field(
        default=None,
        description='An optional VentilationOpening to specify the operable '
        'portion of the Aperture.'
    )


class FaceEnergyPropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^FaceEnergyPropertiesAbridged$') = \
        'FaceEnergyPropertiesAbridged'

    construction: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of an OpaqueConstruction for the Face. If None, the '
        'construction is set by the parent Room construction_set or the '
        'Model global_construction_set.'
    )

    vent_crack: AFNCrack = Field(
        default=None,
        description='An optional AFNCrack to specify airflow through a surface crack '
        'used by the AirflowNetwork.'
    )


class RoomEnergyPropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^RoomEnergyPropertiesAbridged$') = \
        'RoomEnergyPropertiesAbridged'

    construction_set: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of a ConstructionSet to specify all default '
        'constructions for the Faces, Apertures, and Doors of the Room. If '
        'None, the Room will use the Model global_construction_set.'
    )

    program_type: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of a ProgramType to specify all default schedules '
        'and loads for the Room. If None, the Room will have no loads or setpoints.'
    )

    hvac: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='An optional identifier of a HVAC system (such as an IdealAirSystem)'
        ' that specifies how the Room is conditioned. If None, it will be assumed '
        'that the Room is not conditioned.'
    )

    shw: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='An optional identifier of a Service Hot Water (SHW) system '
        'that specifies how the hot water load of the Room is met. If None, the hot '
        'water load will be met with a generic system that only measures thermal load'
        'and does not account for system efficiencies.'
    )

    people: PeopleAbridged = Field(
        default=None,
        description='People object to describe the occupancy of the Room.'
    )

    lighting: LightingAbridged = Field(
        default=None,
        description='Lighting object to describe the lighting usage of the Room.'
    )

    electric_equipment: ElectricEquipmentAbridged = Field(
        default=None,
        description='ElectricEquipment object to describe the electric equipment usage.'
    )

    gas_equipment: GasEquipmentAbridged = Field(
        default=None,
        description='GasEquipment object to describe the gas equipment usage.'
    )

    service_hot_water: ServiceHotWaterAbridged = Field(
        default=None,
        description='ServiceHotWater object to describe the hot water usage.'
    )

    infiltration: InfiltrationAbridged = Field(
        default=None,
        description='Infiltration object to to describe the outdoor air leakage.'
    )

    ventilation: VentilationAbridged = Field(
        default=None,
        description='Ventilation object for the minimum outdoor air requirement.'
    )

    setpoint: SetpointAbridged = Field(
        default=None,
        description='Setpoint object for the temperature setpoints of the Room.'
    )

    daylighting_control: DaylightingControl = Field(
        default=None,
        description='An optional DaylightingControl object to dictate the dimming '
        'of lights. If None, the lighting will respond only to the schedule and '
        'not the daylight conditions within the room.'
    )

    window_vent_control: VentilationControlAbridged = Field(
        default=None,
        description='An optional VentilationControl object to dictate the opening '
        'of windows. If None, the windows will never open.'
    )

    fans: List[VentilationFan] = Field(
        default=None,
        description='An optional list of VentilationFan objects for fans within the '
        'room. Note that these fans are not connected to the heating or cooling system '
        'and are meant to represent the intentional circulation of unconditioned '
        'outdoor air for the purposes of keeping a space cooler, drier or free '
        'of indoor pollutants (as in the case of kitchen or bathroom exhaust fans). '
        'For the specification of mechanical ventilation of conditioned outdoor air, '
        'the Room.ventilation property should be used and the Room should be '
        'given a HVAC that can meet this specification.'
    )

    internal_masses: List[InternalMassAbridged] = Field(
        default=None,
        description='An optional list of of InternalMass objects for thermal mass '
        'exposed to Room air. Note that internal masses assigned this way cannot "see" '
        'solar radiation that may potentially hit them and, as such, caution should be '
        'taken when using this component with internal mass objects that are not '
        'always in shade. Masses are factored into the the thermal calculations '
        'of the Room by undergoing heat transfer with the indoor air.'
    )

    process_loads: List[ProcessAbridged] = Field(
        default=None,
        description='An optional list of of Process objects for process loads within '
        'the room. These can represent kilns, manufacturing equipment, and various '
        'industrial processes. They can also be used to represent wood burning '
        'fireplaces or certain pieces of equipment to be separated from the other '
        'end uses.'
    )


class ModelEnergyProperties(NoExtraBaseModel):

    type: constr(regex='^ModelEnergyProperties$') = \
        'ModelEnergyProperties'

    global_construction_set: GlobalConstructionSet = Field(
        default=GlobalConstructionSet(),
        description='Global Energy construction set.',
        readOnly=True
    )

    construction_sets: List[Union[ConstructionSetAbridged, ConstructionSet]] = Field(
        default=None,
        description='List of all unique ConstructionSets in the Model.'
    )

    constructions: List[
        Union[
            OpaqueConstructionAbridged, WindowConstructionAbridged,
            WindowConstructionShadeAbridged, AirBoundaryConstructionAbridged,
            OpaqueConstruction, WindowConstruction, WindowConstructionShade,
            WindowConstructionDynamicAbridged, WindowConstructionDynamic,
            AirBoundaryConstruction, ShadeConstruction
        ]
    ] = Field(
        default=None,
        description='A list of all unique constructions in the model. This includes '
        'constructions across all Faces, Apertures, Doors, Shades, Room '
        'ConstructionSets, and the global_construction_set.'
    )

    materials: List[
        Union[
            EnergyMaterial, EnergyMaterialNoMass, EnergyMaterialVegetation,
            EnergyWindowMaterialGlazing, EnergyWindowMaterialSimpleGlazSys,
            EnergyWindowMaterialGas, EnergyWindowMaterialGasMixture,
            EnergyWindowMaterialGasCustom, EnergyWindowFrame,
            EnergyWindowMaterialBlind, EnergyWindowMaterialShade
        ]
    ] = Field(
        default=None,
        description='A list of all unique materials in the model. This includes '
        'materials needed to make the Model constructions.'
    )

    hvacs: List[
        Union[
            IdealAirSystemAbridged, VAV, PVAV, PSZ, PTAC, ForcedAirFurnace,
            FCUwithDOASAbridged, WSHPwithDOASAbridged, VRFwithDOASAbridged,
            RadiantwithDOASAbridged,
            FCU, WSHP, VRF, Baseboard, EvaporativeCooler, Residential,
            WindowAC, GasUnitHeater, Radiant,
            DetailedHVAC
        ]
    ] = Field(
        default=None,
        description='List of all unique HVAC systems in the Model.'
    )

    shws: List[SHWSystem] = Field(
        default=None,
        description='List of all unique Service Hot Water (SHW) systems in the Model.'
    )

    program_types: List[Union[ProgramTypeAbridged, ProgramType]] = Field(
        default=None,
        description='List of all unique ProgramTypes in the Model.'
    )

    schedules: List[Union[ScheduleRulesetAbridged, ScheduleFixedIntervalAbridged,
                          ScheduleRuleset, ScheduleFixedInterval]] = Field(
        default=None,
        description='A list of all unique schedules in the model. This includes '
        'schedules across all HVAC systems, ProgramTypes, Rooms, and Shades.'
    )

    schedule_type_limits: List[ScheduleTypeLimit] = Field(
        default=None,
        description='A list of all unique ScheduleTypeLimits in the model. This '
        'all ScheduleTypeLimits needed to make the Model schedules.'
    )

    ventilation_simulation_control: VentilationSimulationControl = Field(
        default=None,
        description='An optional parameter to define the global parameters for '
        'a ventilation cooling.'
    )

    electric_load_center: ElectricLoadCenter = Field(
        default=None,
        description='An optional parameter object that defines the properties '
        'of the model electric loads center that manages on site electricity '
        'generation and conversion.'
    )



"""Detailed HVAC system schema defined using Ironbug."""
from pydantic import Field, constr
from .._base import IDdEnergyBaseModel


class DetailedHVAC(IDdEnergyBaseModel):
    """Detailed HVAC system object defined using IronBug or OpenStudio .NET bindings."""

    type: constr(regex='^DetailedHVAC$') = 'DetailedHVAC'

    specification: dict = Field(
        ...,
        description='A JSON-serializable dictionary representing the full '
        'specification of the detailed system. This can be obtained by calling '
        'the ToJson() method on any IronBug HVAC system and then serializing '
        'the resulting JSON string into a Python dictionary using the native '
        'Python json package. Note that the Rooms that the HVAC is assigned to '
        'must be specified as ThermalZones under this specification in order '
        'for the resulting Model this HVAC is a part of to be valid.'
    )



"""Base class used by various schema objects."""
from pydantic import Field
import datetime

from .._base import NoExtraBaseModel


class EnergyBaseModel(NoExtraBaseModel):
    """Base class for all objects requiring a valid EnergyPlus identifier."""

    identifier: str = Field(
        ...,
        regex=r'^[^,;!\n\t]+$',
        min_length=1,
        max_length=100,
        description='Text string for a unique object ID. This identifier remains '
        'constant as the object is mutated, copied, and serialized to different '
        'formats (eg. dict, idf, osm). This identifier is also used to reference '
        'the object across a Model. It must be < 100 characters, use only '
        'ASCII characters and exclude (, ; ! \\n \\t).'
    )

    display_name: str = Field(
        default=None,
        description='Display name of the object with no character restrictions.'
    )


class IDdEnergyBaseModel(EnergyBaseModel):
    """Base class for all objects requiring an EnergyPlus identifier and user_data."""

    user_data: dict = Field(
        default=None,
        description='Optional dictionary of user data associated with the object.'
        'All keys and values of this dictionary should be of a standard data '
        'type to ensure correct serialization of the object (eg. str, float, '
        'int, list).'
    )


class DatedBaseModel(NoExtraBaseModel):
    """Base class for all objects needing to check for a valid Date."""

    @staticmethod
    def check_date(v, leap_pear_override=None):
        """Ensure valid date.

        Args:
            v: The Date array to be validated.
            leap_pear_override: Boolean to override the typical check for
                leap year serialization of date arrays. If not None, this
                boolean wil dictate whether the date is a leap year or not.
        """
        if (len(v) == 3 and v[2]) or leap_pear_override:
            try:
                datetime.date(2016, v[0], v[1])
            except ValueError:
                raise ValueError('{}/{} is not a valid date.'.format(v[0], v[1]))
        else:
            try:
                datetime.date(2017, v[0], v[1])
            except ValueError:
                raise ValueError('{}/{} is not a valid date.'.format(v[0], v[1]))
        return v
